from __future__ import annotations

from typing import Literal

import numpy as np
from typing_extensions import assert_never


def cv2_imencode_with_fallback(
    format: Literal["png", "jpeg"],
    image: np.ndarray,
    jpeg_quality: int | None,
    channel_ordering: Literal["rgb", "bgr"],
) -> bytes:
    """Helper for encoding images to bytes using OpenCV or imageio.

    We default to OpenCV if available, which we find is usually faster:
        https://github.com/viser-project/viser/pull/494

    We fall back to imageio if OpenCV is not available. This lets us avoid
    adding OpenCV as a strict dependency, since it can be annoying to install
    on some machines:
        https://github.com/viser-project/viser/issues/535
    """
    if jpeg_quality is None:
        jpeg_quality = 75  # Default JPEG quality if not specified.

    try:
        import cv2
    except ImportError:
        # Fall back to imageio if cv2 is not available.
        import imageio.v3 as iio

        if channel_ordering == "bgr":
            # Convert to BGR if needed.
            image = image[
                :, :, np.array((2, 1, 0, 3) if image.shape[-1] == 4 else (2, 1, 0))
            ]
        return (
            iio.imwrite("<bytes>", image, extension=".jpeg", quality=jpeg_quality)
            if format == "jpeg"
            else iio.imwrite("<bytes>", image, extension=".png")
        )

    # OpenCV is available!
    if channel_ordering == "rgb":
        # Convert to BGR if needed.
        image = image[
            :, :, np.array((2, 1, 0, 3) if image.shape[-1] == 4 else (2, 1, 0))
        ]
    if format == "png":
        success, encoded_image = cv2.imencode(".png", image)
    elif format == "jpeg":
        if jpeg_quality is not None:
            success, encoded_image = cv2.imencode(
                ".jpeg", image, [int(cv2.IMWRITE_JPEG_QUALITY), jpeg_quality]
            )
        else:
            success, encoded_image = cv2.imencode(format, image)
    else:
        assert_never(format)

    if not success:
        raise RuntimeError("Failed to encode image.")

    return encoded_image.tobytes()



from __future__ import annotations

import abc
from functools import cached_property
from typing import Any, Dict, Generic, Protocol, TypeVar, get_type_hints

import numpy as np
import numpy.typing as npt

# Type variable for props


class HasProps(Protocol):
    props: Any  # One of the `*Props` objects in _messages.py.


TImpl = TypeVar("TImpl", bound=HasProps)


def colors_to_uint8(colors: np.ndarray) -> npt.NDArray[np.uint8]:
    """Convert intensity values to uint8. We assume the range [0,1] for floats, and
    [0,255] for integers. Accepts any shape."""
    if colors.dtype != np.uint8:
        if np.issubdtype(colors.dtype, np.floating):
            colors = np.clip(colors * 255.0, 0, 255).astype(np.uint8)
        if np.issubdtype(colors.dtype, np.integer):
            colors = np.clip(colors, 0, 255).astype(np.uint8)
    return colors


class AssignablePropsBase(Generic[TImpl]):
    """Base class for all API objects with assignable properties."""

    _impl: TImpl

    def __init__(self, impl: TImpl):
        # Make sure arrays are copied to avoid shared references.
        # This will also make sure that our `np.array_equal` checks below work
        # correctly.
        for k, v in vars(impl.props).items():
            if isinstance(v, np.ndarray):
                setattr(impl.props, k, v.copy())

        # Store the implementation object.
        self._impl = impl

    def _cast_value_recursive(self, hint: Any, value: Any, prop_name: str) -> Any:
        """Recursively cast values to match type hints, handling arrays and tuples."""
        # Handle numpy arrays
        if hint == npt.NDArray[np.float16]:
            return np.asarray(value).astype(np.float16)
        elif hint == npt.NDArray[np.float32]:
            return np.asarray(value).astype(np.float32)
        elif hint == npt.NDArray[np.float64]:
            return np.asarray(value).astype(np.float64)
        elif hint == npt.NDArray[np.uint8] and "color" in prop_name:
            return colors_to_uint8(value)
        if isinstance(value, np.ndarray):
            return value

        # Handle tuple[T, ...] pattern
        if (
            isinstance(value, tuple)
            and hasattr(hint, "__origin__")
            and hint.__origin__ is tuple
            and hasattr(hint, "__args__")
            and len(hint.__args__) == 2
            and hint.__args__[1] is ...
        ):
            element_type = hint.__args__[0]
            return tuple(
                self._cast_value_recursive(element_type, item, prop_name)
                for item in value
            )

        return value

    def _cast_array_dtypes(
        self, prop_hints: Dict[str, Any], prop_name: str, value: np.ndarray
    ) -> np.ndarray:
        """Helper to cast array values to the correct data type."""
        return self._cast_value_recursive(prop_hints[prop_name], value, prop_name)

    @cached_property
    def _prop_hints(self) -> Dict[str, Any]:
        return get_type_hints(type(self._impl.props))

    @abc.abstractmethod
    def _queue_update(self, name: str, value: Any) -> None:
        """Queue an update message with the property change."""


def props_setattr(self, name: str, value: Any) -> None:
    if name == "_impl":
        return object.__setattr__(self, name, value)

    # If it's a property with a setter, use the setter.
    prop = getattr(self.__class__, name, None)
    if isinstance(prop, property) and prop.fset is not None:
        prop.fset(self, value)
        return

    # Try to handle as a props field.
    if name in self._prop_hints:
        # Handle type casting (arrays, tuples of arrays, etc.).
        value = self._cast_value_recursive(self._prop_hints[name], value, name)
        current_value = getattr(self._impl.props, name)

        # Skip update if value hasn't changed.
        try:
            hash(current_value)
            if current_value == value:
                return
        except (TypeError, ValueError):
            pass

        # Update the value based on type.
        if isinstance(value, np.ndarray):
            if hasattr(current_value, "dtype"):
                # Ensure consistent dtype.
                if value.dtype != current_value.dtype:
                    value = value.astype(current_value.dtype)
                if np.array_equal(current_value, value):
                    return

            # In-place update for same shape arrays.
            if hasattr(current_value, "shape") and value.shape == current_value.shape:
                current_value[:] = value
            else:
                setattr(self._impl.props, name, value.copy())
        else:
            # Non-array properties
            setattr(self._impl.props, name, value)
    else:
        return object.__setattr__(self, name, value)

    self._queue_update(name, value)


def props_getattr(self, name: str) -> Any:
    if name in self._prop_hints:
        return getattr(self._impl.props, name)
    else:
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )


AssignablePropsBase.__setattr__ = props_setattr  # type: ignore
AssignablePropsBase.__getattr__ = props_getattr  # type: ignore



from __future__ import annotations

import dataclasses
from typing import Any, Literal

from typing_extensions import override

from viser._assignable_props_api import AssignablePropsBase

from ._messages import NotificationMessage, NotificationProps, RemoveNotificationMessage
from .infra._infra import WebsockClientConnection


@dataclasses.dataclass
class _NotificationHandleState:
    websock_interface: WebsockClientConnection
    uuid: str
    props: NotificationProps


class NotificationHandle(
    NotificationProps, AssignablePropsBase[_NotificationHandleState]
):
    """Handle for a notification in our visualizer."""

    def __init__(self, impl: _NotificationHandleState) -> None:
        self._impl = impl

    @override
    def _queue_update(self, name: str, value: Any) -> None:
        """Queue an update message with the property change."""
        # For notifications, we'll just send the whole props object when a
        # property is reassigned. Deduplication in the message buffer will
        # debounce this when multiple properties are updated in succession.
        del name, value
        self._sync_with_client("update")

    def _sync_with_client(self, mode: Literal["show", "update"]) -> None:
        msg = NotificationMessage(mode, self._impl.uuid, self._impl.props)
        self._impl.websock_interface.queue_message(msg)

    def remove(self) -> None:
        self._impl.websock_interface.get_message_buffer().remove_from_buffer(
            # Don't send outdated GUI updates to new clients.
            # This is brittle...
            lambda message: getattr(message, "uuid", None) == self._impl.uuid
        )
        msg = RemoveNotificationMessage(self._impl.uuid)
        self._impl.websock_interface.queue_message(msg)



"""Configuration types for uPlot charts in viser.

For comprehensive documentation about uPlot options, see:
https://github.com/leeoniya/uPlot/tree/1.6.32/docs
"""

from ._uplot_types import Axis as Axis
from ._uplot_types import Axis_Border as Axis_Border
from ._uplot_types import Axis_Grid as Axis_Grid
from ._uplot_types import Axis_Ticks as Axis_Ticks
from ._uplot_types import Band as Band
from ._uplot_types import Cursor as Cursor
from ._uplot_types import Cursor_Bind as Cursor_Bind
from ._uplot_types import Cursor_Drag as Cursor_Drag
from ._uplot_types import Cursor_Focus as Cursor_Focus
from ._uplot_types import Cursor_Hover as Cursor_Hover
from ._uplot_types import Cursor_Points as Cursor_Points
from ._uplot_types import Cursor_Sync as Cursor_Sync
from ._uplot_types import Focus as Focus
from ._uplot_types import Legend as Legend
from ._uplot_types import Legend_Markers as Legend_Markers
from ._uplot_types import Scale as Scale
from ._uplot_types import Select as Select
from ._uplot_types import Series as Series
from ._uplot_types import Series_Facet as Series_Facet
from ._uplot_types import Series_Points as Series_Points



import zipfile
from functools import lru_cache
from pathlib import Path

from ._icons_enum import IconName

ICONS_DIR = Path(__file__).absolute().parent / "_icons"


@lru_cache(maxsize=32)
def svg_from_icon(icon_name: IconName) -> str:
    """Read an icon and return it as a UTF string; we expect this to be an
    <svg /> tag."""
    assert isinstance(icon_name, str)
    icons_zipfile = ICONS_DIR / "tabler-icons.zip"

    with zipfile.ZipFile(icons_zipfile) as zip_file:
        with zip_file.open(f"{icon_name}.svg") as icon_file:
            out = icon_file.read()

    return out.decode("utf-8")



from __future__ import annotations

import base64
import dataclasses
import re
import time
import uuid
import warnings
from collections.abc import Coroutine
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Literal,
    Tuple,
    TypeVar,
    overload,
)

import imageio.v3 as iio
import numpy as np
from typing_extensions import Protocol, override

from ._assignable_props_api import AssignablePropsBase
from ._icons import svg_from_icon
from ._icons_enum import IconName
from ._messages import (
    GuiBaseProps,
    GuiButtonGroupProps,
    GuiButtonProps,
    GuiCheckboxProps,
    GuiCloseModalMessage,
    GuiDropdownProps,
    GuiFolderProps,
    GuiHtmlProps,
    GuiImageProps,
    GuiMarkdownProps,
    GuiMultiSliderProps,
    GuiNumberProps,
    GuiPlotlyProps,
    GuiProgressBarProps,
    GuiRemoveMessage,
    GuiRgbaProps,
    GuiRgbProps,
    GuiSliderProps,
    GuiTabGroupProps,
    GuiTextProps,
    GuiUpdateMessage,
    GuiUploadButtonProps,
    GuiUplotProps,
    GuiVector2Props,
    GuiVector3Props,
)
from ._scene_api import _encode_image_binary
from .infra import ClientId

if TYPE_CHECKING:
    import plotly.graph_objects as go

    from ._gui_api import GuiApi
    from ._viser import ClientHandle


T = TypeVar("T")
TGuiHandle = TypeVar("TGuiHandle", bound="_GuiInputHandle")
NoneOrCoroutine = TypeVar("NoneOrCoroutine", None, Coroutine)


def _make_uuid() -> str:
    """Return a unique ID for referencing GUI elements."""
    return str(uuid.uuid4())


class GuiContainerProtocol(Protocol):
    _children: dict[str, SupportsRemoveProtocol] = dataclasses.field(
        default_factory=dict
    )


class SupportsRemoveProtocol(Protocol):
    def remove(self) -> None: ...


class GuiPropsProtocol(Protocol):
    order: float


@dataclasses.dataclass
class _GuiHandleState(Generic[T]):
    """Internal API for GUI elements."""

    uuid: str
    gui_api: GuiApi
    value: T
    props: GuiPropsProtocol
    parent_container_id: str
    """Container that this GUI input was placed into."""

    update_timestamp: float = 0.0
    update_cb: list[Callable[[GuiEvent], None | Coroutine]] = dataclasses.field(
        default_factory=list
    )
    """Registered functions to call when this input is updated."""

    is_button: bool = False
    """Indicates a button element, which requires special handling."""

    sync_cb: Callable[[ClientId, dict[str, Any]], None] | None = None
    """Callback for synchronizing inputs across clients."""

    removed: bool = False


@dataclasses.dataclass
class _GuiButtonHandleState(_GuiHandleState[bool]):
    """Internal API for button GUI elements with hold callback support."""

    hold_cbs_from_freq: dict[float, list[Callable[[GuiEvent], None | Coroutine]]] = (
        dataclasses.field(default_factory=dict)
    )
    """Mapping from frequency (Hz) to list of callbacks to call when button is held."""


# Not exported for now because some GUI handles don't currently inhert from
# `_GuiHandle`: notably `GuiModalHandle` and `GuiTabHandle`. These would fail
# isinstance checks, which would be confusing!
class _GuiHandle(Generic[T], AssignablePropsBase[_GuiHandleState]):
    def __init__(self, impl: _GuiHandleState[T]) -> None:
        super().__init__(impl=impl)
        parent = self._impl.gui_api._container_handle_from_uuid[
            self._impl.parent_container_id
        ]
        parent._children[self._impl.uuid] = self

        if isinstance(self, _GuiInputHandle):
            self._impl.gui_api._gui_input_handle_from_uuid[self._impl.uuid] = self

    @override
    def _queue_update(self, name: str, value: Any) -> None:
        self._impl.gui_api._websock_interface.queue_message(
            GuiUpdateMessage(self._impl.uuid, {name: value})
        )

    def remove(self) -> None:
        """Permanently remove this GUI element from the visualizer."""

        # Warn if already removed.
        if self._impl.removed:
            warnings.warn(
                f"Attempted to remove an already removed {self.__class__.__name__}.",
                stacklevel=2,
            )
            return
        self._impl.removed = True

        # Send remove to client(s) + update internal state.
        gui_api = self._impl.gui_api
        gui_api._websock_interface.get_message_buffer().remove_from_buffer(
            # Don't send outdated GUI updates to new clients.
            # This is brittle...
            lambda message: getattr(message, "uuid", None) == self._impl.uuid
        )
        gui_api._websock_interface.queue_message(GuiRemoveMessage(self._impl.uuid))
        parent = gui_api._container_handle_from_uuid[self._impl.parent_container_id]
        parent._children.pop(self._impl.uuid)

        if isinstance(self, _GuiInputHandle):
            gui_api._gui_input_handle_from_uuid.pop(self._impl.uuid)


class _GuiInputHandle(
    _GuiHandle[T],
    Generic[T],
    GuiBaseProps,
):
    @property
    def value(self) -> T:
        """Value of the GUI input. Synchronized automatically when assigned.

        :meta private:
        """
        # ^Note: we mark this property as private for Sphinx because I haven't
        # been able to get it to resolve the TypeVar in a readable way.
        # For the documentation's sake, we'll be manually adding ::attribute directives below.
        return self._impl.value

    @value.setter
    def value(self, value: T | np.ndarray) -> None:
        if isinstance(value, np.ndarray):
            assert len(value.shape) <= 1, f"{value.shape} should be at most 1D!"
            value = tuple(map(float, value))  # type: ignore

        # Send to client, except for buttons.
        if not self._impl.is_button:
            self._impl.gui_api._websock_interface.queue_message(
                GuiUpdateMessage(self._impl.uuid, {"value": value})
            )

        # Set internal state. We automatically convert numpy arrays to the expected
        # internal type. (eg 1D arrays to tuples)
        self._impl.value = type(self._impl.value)(value)  # type: ignore
        self._impl.update_timestamp = time.time()

        # Call update callbacks.
        for cb in self._impl.update_cb:
            # As a design decision: we choose to call update callbacks
            # synchronously instead of in the thread pool. It's rare that there
            # are significant blocking callbacks for GUI updates; this also
            # reduces the likelihood of many common race conditions.
            cb_out = cb(GuiEvent(client_id=None, client=None, target=self))
            if isinstance(cb_out, Coroutine):
                self._impl.gui_api._event_loop.create_task(cb_out)

    @property
    def update_timestamp(self) -> float:
        """Read-only timestamp when this input was last updated."""
        return self._impl.update_timestamp


StringType = TypeVar("StringType", bound=str)


# GuiInputHandle[T] is used for all inputs except for buttons.
#
# We inherit from _GuiInputHandle to special-case buttons because the usage semantics
# are slightly different: we have `on_click()` instead of `on_update()`.
class GuiInputHandle(_GuiInputHandle[T], Generic[T]):
    """A handle is created for each GUI element that is added in `viser`.
    Handles can be used to read and write state.

    When a GUI element is added via :attr:`ViserServer.gui`, state is
    synchronized between all connected clients. When a GUI element is added via
    :attr:`ClientHandle.gui`, state is local to a specific client.
    """

    def on_update(
        self: TGuiHandle, func: Callable[[GuiEvent[TGuiHandle]], NoneOrCoroutine]
    ) -> Callable[[GuiEvent[TGuiHandle]], NoneOrCoroutine]:
        """Attach a function to call when a GUI input is updated.

        Note:
        - If `func` is a regular function (defined with `def`), it will be executed in a thread pool.
        - If `func` is an async function (defined with `async def`), it will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._impl.update_cb.append(func)
        return func

    def remove_update_callback(
        self, callback: Literal["all"] | Callable = "all"
    ) -> None:
        """Remove update callbacks from the GUI input.

        Args:
            callback: Either "all" to remove all callbacks, or a specific callback function to remove.
        """
        if callback == "all":
            self._impl.update_cb.clear()
        else:
            self._impl.update_cb = [cb for cb in self._impl.update_cb if cb != callback]


class GuiCheckboxHandle(GuiInputHandle[bool], GuiCheckboxProps):
    """Handle for checkbox inputs.

    .. attribute:: value
       :type: bool

       Value of the input. Synchronized automatically when assigned.
    """


class GuiTextHandle(GuiInputHandle[str], GuiTextProps):
    """Handle for text inputs.

    .. attribute:: value
       :type: str

       Value of the input. Synchronized automatically when assigned.
    """


IntOrFloat = TypeVar("IntOrFloat", int, float)


class GuiNumberHandle(GuiInputHandle[IntOrFloat], Generic[IntOrFloat], GuiNumberProps):
    """Handle for number inputs.

    .. attribute:: value
       :type: IntOrFloat

       Value of the input. Synchronized automatically when assigned.
    """


class GuiSliderHandle(GuiInputHandle[IntOrFloat], Generic[IntOrFloat], GuiSliderProps):
    """Handle for slider inputs.

    .. attribute:: value
       :type: IntOrFloat

       Value of the input. Synchronized automatically when assigned.
    """


class GuiMultiSliderHandle(
    GuiInputHandle[Tuple[IntOrFloat, ...]], Generic[IntOrFloat], GuiMultiSliderProps
):
    """Handle for multi-slider inputs.

    .. attribute:: value
       :type: tuple[IntOrFloat, ...]

       Value of the input. Synchronized automatically when assigned.
    """


class GuiRgbHandle(GuiInputHandle[Tuple[int, int, int]], GuiRgbProps):
    """Handle for RGB color inputs.

    .. attribute:: value
       :type: tuple[int, int, int]

       Value of the input. Synchronized automatically when assigned.
    """


class GuiRgbaHandle(GuiInputHandle[Tuple[int, int, int, int]], GuiRgbaProps):
    """Handle for RGBA color inputs.

    .. attribute:: value
       :type: tuple[int, int, int, int]

       Value of the input. Synchronized automatically when assigned.
    """


class GuiVector2Handle(GuiInputHandle[Tuple[float, float]], GuiVector2Props):
    """Handle for 2D vector inputs.

    .. attribute:: value
       :type: tuple[float, float]

       Value of the input. Synchronized automatically when assigned.
    """


class GuiVector3Handle(GuiInputHandle[Tuple[float, float, float]], GuiVector3Props):
    """Handle for 3D vector inputs.

    .. attribute:: value
       :type: tuple[float, float, float]

       Value of the input. Synchronized automatically when assigned.
    """


@dataclasses.dataclass(frozen=True)
class GuiEvent(Generic[TGuiHandle]):
    """Information associated with a GUI event, such as an update or click.

    Passed as input to callback functions."""

    client: ClientHandle | None
    """Client that triggered this event."""
    client_id: int | None
    """ID of client that triggered this event."""
    target: TGuiHandle
    """GUI element that was affected."""


class GuiButtonHandle(_GuiInputHandle[bool], GuiButtonProps):
    """Handle for a button input in our visualizer.

    .. attribute:: value
       :type: bool

       Value of the button. Set to `True` when the button is pressed. Can be manually set back to `False`.
    """

    def __init__(self, _impl: _GuiButtonHandleState, _icon: IconName | None):
        super().__init__(impl=_impl)
        self._icon = _icon

    @property
    def _button_impl(self) -> _GuiButtonHandleState:
        """Access the button-specific implementation state."""
        assert isinstance(self._impl, _GuiButtonHandleState)
        return self._impl

    @property
    def icon(self) -> IconName | None:
        """Icon to display on the button. When set to None, no icon is displayed."""
        return self._icon

    @icon.setter
    def icon(self, icon: IconName | None) -> None:
        self._icon = icon
        self._icon_html = None if icon is None else svg_from_icon(icon)

    def on_click(
        self: TGuiHandle, func: Callable[[GuiEvent[TGuiHandle]], NoneOrCoroutine]
    ) -> Callable[[GuiEvent[TGuiHandle]], NoneOrCoroutine]:
        """Attach a function to call when a button is pressed.

        Note:
        - If `func` is a regular function (defined with `def`), it will be executed in a thread pool.
        - If `func` is an async function (defined with `async def`), it will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._impl.update_cb.append(func)
        return func

    # Type alias for button hold callbacks.
    _HoldCallback = Callable[["GuiEvent[GuiButtonHandle]"], "None | Coroutine"]

    @overload
    def on_hold(
        self,
        func: None = None,
        callback_hz: float = 10.0,
    ) -> Callable[[_HoldCallback], _HoldCallback]: ...

    @overload
    def on_hold(
        self,
        func: _HoldCallback,
        callback_hz: float = 10.0,
    ) -> _HoldCallback: ...

    def on_hold(
        self,
        func: _HoldCallback | None = None,
        callback_hz: float = 10.0,
    ) -> Callable[[_HoldCallback], _HoldCallback] | _HoldCallback:
        """Attach a function to call repeatedly while a button is held down.

        The callback will be triggered immediately when the button is pressed,
        and then repeatedly at the specified frequency until released.

        Can be used as a decorator with or without arguments:
            @button.on_hold
            def callback(event): ...

            @button.on_hold(callback_hz=30.0)
            def callback(event): ...

        Or called directly:
            button.on_hold(callback)
            button.on_hold(callback, callback_hz=30.0)

        Args:
            func: The callback function to attach. If None, returns a decorator.
            callback_hz: The frequency in Hz at which to call the callback while
                the button is held. Defaults to 10.0 Hz.

        Note:
        - If `func` is a regular function (defined with `def`), it will be executed in a thread pool.
        - If `func` is an async function (defined with `async def`), it will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        button_impl = self._button_impl

        def register_callback(
            f: GuiButtonHandle._HoldCallback,
        ) -> GuiButtonHandle._HoldCallback:
            # Add callback to the frequency-specific list.
            if callback_hz not in button_impl.hold_cbs_from_freq:
                button_impl.hold_cbs_from_freq[callback_hz] = []
            button_impl.hold_cbs_from_freq[callback_hz].append(f)

            # Update the prop to notify client of new frequency.
            self._hold_callback_freqs = tuple(button_impl.hold_cbs_from_freq.keys())

            return f

        if func is not None:
            return register_callback(func)
        return register_callback


@dataclasses.dataclass
class UploadedFile:
    """Result of a file upload."""

    name: str
    """Name of the file."""
    content: bytes
    """Contents of the file."""


class GuiUploadButtonHandle(_GuiInputHandle[UploadedFile], GuiUploadButtonProps):
    """Handle for an upload file button in our visualizer.

    The `.value` attribute will be updated with the contents of uploaded files.

    .. attribute:: value
       :type: UploadedFile

       Value of the input. Contains information about the uploaded file.
    """

    def __init__(self, _impl: _GuiHandleState[UploadedFile], _icon: IconName | None):
        super().__init__(impl=_impl)
        self._icon = _icon

    @property
    def icon(self) -> IconName | None:
        """Icon to display on the upload button. When set to None, no icon is displayed."""
        return self._icon

    @icon.setter
    def icon(self, icon: IconName | None) -> None:
        self._icon = icon
        self._icon_html = None if icon is None else svg_from_icon(icon)

    def on_upload(
        self: TGuiHandle, func: Callable[[GuiEvent[TGuiHandle]], NoneOrCoroutine]
    ) -> Callable[[GuiEvent[TGuiHandle]], NoneOrCoroutine]:
        """Attach a function to call when a file is uploaded.

        Note:
        - If `func` is a regular function (defined with `def`), it will be executed in a thread pool.
        - If `func` is an async function (defined with `async def`), it will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._impl.update_cb.append(func)
        return func


class GuiButtonGroupHandle(_GuiInputHandle[str], GuiButtonGroupProps):
    """Handle for a button group input in our visualizer.

    .. attribute:: value
       :type: str

       Value of the input. Represents the currently selected button in the group.
    """

    def on_click(
        self: TGuiHandle, func: Callable[[GuiEvent[TGuiHandle]], NoneOrCoroutine]
    ) -> Callable[[GuiEvent[TGuiHandle]], NoneOrCoroutine]:
        """Attach a function to call when a button in the group is clicked.

        Note:
        - If `func` is a regular function (defined with `def`), it will be executed in a thread pool.
        - If `func` is an async function (defined with `async def`), it will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._impl.update_cb.append(func)
        return func

    @property
    def disabled(self) -> bool:
        """Button groups cannot be disabled."""
        return False

    @disabled.setter
    def disabled(self, disabled: bool) -> None:  # type: ignore
        """Button groups cannot be disabled."""
        assert not disabled, "Button groups cannot be disabled."


class GuiDropdownHandle(
    GuiInputHandle[StringType], Generic[StringType], GuiDropdownProps
):
    """Handle for a dropdown-style GUI input in our visualizer.

    .. attribute:: value
       :type: StringType

       Value of the input. Represents the currently selected option in the dropdown.
    """

    @property
    def options(self) -> tuple[StringType, ...]:
        """Options for our dropdown. Synchronized automatically when assigned.

        For projects that care about typing: the static type of `options` should be
        consistent with the `StringType` associated with a handle. Literal types will be
        inferred where possible when handles are instantiated; for the most flexibility,
        we can declare handles as `GuiDropdownHandle[str]`.
        """
        assert isinstance(self._impl.props, GuiDropdownProps)
        return self._impl.props.options  # type: ignore

    @options.setter
    def options(self, options: Iterable[StringType]) -> None:  # type: ignore
        assert isinstance(self._impl.props, GuiDropdownProps)
        options = tuple(options)
        self._impl.props.options = options

        self._impl.gui_api._websock_interface.queue_message(
            GuiUpdateMessage(
                self._impl.uuid,
                {"options": options},
            )
        )
        if self.value not in options:
            self.value = options[0]


class GuiTabGroupHandle(_GuiHandle[None], GuiTabGroupProps):
    """Handle for a tab group. Call :meth:`add_tab()` to add a tab."""

    def __init__(self, _impl: _GuiHandleState[None]) -> None:
        super().__init__(impl=_impl)
        self._tab_handles: list[GuiTabHandle] = []

    def add_tab(self, label: str, icon: IconName | None = None) -> GuiTabHandle:
        """Add a tab. Returns a handle we can use to add GUI elements to it."""

        uuid = _make_uuid()

        # We may want to make this thread-safe in the future.
        out = GuiTabHandle(_parent=self, _id=uuid, _label=label, _icon=icon)

        self._tab_handles.append(out)
        self._tab_labels = self._tab_labels + (label,)
        self._tab_icons_html = self._tab_icons_html + (
            None if icon is None else svg_from_icon(icon),
        )
        self._tab_container_ids = tuple(handle._id for handle in self._tab_handles)
        return out

    def __post_init__(self) -> None:
        parent = self._impl.gui_api._container_handle_from_uuid[
            self._impl.parent_container_id
        ]
        parent._children[self._impl.uuid] = self

    def remove(self) -> None:
        """Remove this tab group and all contained GUI elements."""
        # Warn if already removed.
        if self._impl.removed:
            warnings.warn(
                f"Attempted to remove an already removed {self.__class__.__name__}.",
                stacklevel=2,
            )
            return
        self._impl.removed = True

        # Remove tabs, then self.
        for tab in tuple(self._tab_handles):
            tab.remove()
        gui_api = self._impl.gui_api
        gui_api._websock_interface.get_message_buffer().remove_from_buffer(
            # Don't send outdated GUI updates to new clients.
            lambda message: (
                isinstance(message, GuiUpdateMessage)
                and message.uuid == self._impl.uuid
            )
        )
        gui_api._websock_interface.queue_message(GuiRemoveMessage(self._impl.uuid))
        parent = gui_api._container_handle_from_uuid[self._impl.parent_container_id]
        parent._children.pop(self._impl.uuid)


@dataclasses.dataclass
class GuiTabHandle:
    """Use as a context to place GUI elements into a tab."""

    _parent: GuiTabGroupHandle
    _id: str  # Used as container ID of children.
    _label: str
    _icon: IconName | None
    _container_id_restore: str | None = None
    _children: dict[str, SupportsRemoveProtocol] = dataclasses.field(
        default_factory=dict
    )
    _removed: bool = False

    @property
    def icon(self) -> IconName | None:
        """Icon to display on the tab. When set to None, no icon is displayed."""
        return self._icon

    @icon.setter
    def icon(self, icon: IconName | None) -> None:
        self._icon = icon
        # Find the index of this tab in the parent's tab list.
        tab_index = self._parent._tab_handles.index(self)
        # Update the icon HTML in the parent's tuple.
        icons_list = list(self._parent._tab_icons_html)
        icons_list[tab_index] = None if icon is None else svg_from_icon(icon)
        self._parent._tab_icons_html = tuple(icons_list)

    def __enter__(self) -> GuiTabHandle:
        self._container_id_restore = self._parent._impl.gui_api._get_container_uuid()
        self._parent._impl.gui_api._set_container_uuid(self._id)
        return self

    def __exit__(self, *args) -> None:
        del args
        assert self._container_id_restore is not None
        self._parent._impl.gui_api._set_container_uuid(self._container_id_restore)
        self._container_id_restore = None

    def __post_init__(self) -> None:
        self._parent._impl.gui_api._container_handle_from_uuid[self._id] = self

    def remove(self) -> None:
        """Permanently remove this tab and all contained GUI elements from the
        visualizer."""
        # Warn if already removed.
        if self._removed:
            warnings.warn(
                f"Attempted to remove an already removed {self.__class__.__name__}.",
                stacklevel=2,
            )
            return
        self._removed = True

        # We may want to make this thread-safe in the future.
        found_index = -1
        for i, tab in enumerate(self._parent._tab_handles):
            if tab is self:
                found_index = i
                break
        assert found_index != -1, "Tab already removed!"

        self._parent._tab_labels = (
            self._parent._tab_labels[:found_index]
            + self._parent._tab_labels[found_index + 1 :]
        )
        self._parent._tab_icons_html = (
            self._parent._tab_icons_html[:found_index]
            + self._parent._tab_icons_html[found_index + 1 :]
        )
        self._parent._tab_handles = (
            self._parent._tab_handles[:found_index]
            + self._parent._tab_handles[found_index + 1 :]
        )

        for child in tuple(self._children.values()):
            child.remove()
        self._parent._impl.gui_api._container_handle_from_uuid.pop(self._id)


class GuiFolderHandle(_GuiHandle[None], GuiFolderProps):
    """Use as a context to place GUI elements into a folder."""

    def __init__(self, _impl: _GuiHandleState[None]) -> None:
        super().__init__(impl=_impl)
        self._impl.gui_api._container_handle_from_uuid[self._impl.uuid] = self
        self._children = {}
        parent = self._impl.gui_api._container_handle_from_uuid[
            self._impl.parent_container_id
        ]
        parent._children[self._impl.uuid] = self

    def __enter__(self) -> GuiFolderHandle:
        self._container_id_restore = self._impl.gui_api._get_container_uuid()
        self._impl.gui_api._set_container_uuid(self._impl.uuid)
        return self

    def __exit__(self, *args) -> None:
        del args
        assert self._container_id_restore is not None
        self._impl.gui_api._set_container_uuid(self._container_id_restore)
        self._container_id_restore = None

    def remove(self) -> None:
        """Permanently remove this folder and all contained GUI elements from the
        visualizer."""
        # Warn if already removed.
        if self._impl.removed:
            warnings.warn(
                f"Attempted to remove an already removed {self.__class__.__name__}.",
                stacklevel=2,
            )
            return
        self._impl.removed = True

        # Remove children, then self.
        gui_api = self._impl.gui_api
        gui_api._websock_interface.get_message_buffer().remove_from_buffer(
            # Don't send outdated GUI updates to new clients.
            lambda message: (
                isinstance(message, GuiUpdateMessage)
                and message.uuid == self._impl.uuid
            )
        )
        gui_api._websock_interface.queue_message(GuiRemoveMessage(self._impl.uuid))
        for child in tuple(self._children.values()):
            child.remove()
        parent = gui_api._container_handle_from_uuid[self._impl.parent_container_id]
        parent._children.pop(self._impl.uuid)
        gui_api._container_handle_from_uuid.pop(self._impl.uuid)


@dataclasses.dataclass
class GuiModalHandle:
    """Use as a context to place GUI elements into a modal."""

    _gui_api: GuiApi
    _uuid: str  # Used as container ID of children.
    _container_uuid_restore: str | None = None
    _children: dict[str, SupportsRemoveProtocol] = dataclasses.field(
        default_factory=dict
    )

    def __enter__(self) -> GuiModalHandle:
        self._container_uuid_restore = self._gui_api._get_container_uuid()
        self._gui_api._set_container_uuid(self._uuid)
        return self

    def __exit__(self, *args) -> None:
        del args
        assert self._container_uuid_restore is not None
        self._gui_api._set_container_uuid(self._container_uuid_restore)
        self._container_uuid_restore = None

    def __post_init__(self) -> None:
        self._gui_api._container_handle_from_uuid[self._uuid] = self
        self._gui_api._modal_handle_from_uuid[self._uuid] = self

    def close(self) -> None:
        """Close this modal and permananently remove all contained GUI elements."""
        self._gui_api._websock_interface.queue_message(
            GuiCloseModalMessage(self._uuid),
        )
        for child in tuple(self._children.values()):
            child.remove()
        self._gui_api._container_handle_from_uuid.pop(self._uuid)
        self._gui_api._modal_handle_from_uuid.pop(self._uuid)


def _get_data_url(url: str, image_root: Path | None) -> str:
    if not url.startswith("http") and not image_root:
        warnings.warn(
            (
                "No `image_root` provided. All relative paths will be scoped to viser's"
                " installation path."
            ),
            stacklevel=2,
        )
    if url.startswith("http") or url.startswith("data:"):
        return url
    if image_root is None:
        image_root = Path(__file__).parent
    try:
        image = iio.imread(image_root / url)
        _, binary = _encode_image_binary(image, "png")
        url = base64.b64encode(binary).decode("utf-8")
        return f"data:image/png;base64,{url}"
    except (IOError, FileNotFoundError):
        warnings.warn(
            f"Failed to read image {url}, with image_root set to {image_root}.",
            stacklevel=2,
        )
        return url


def _parse_markdown(markdown: str, image_root: Path | None) -> str:
    markdown = re.sub(
        r"\!\[([^]]*)\]\(([^]]*)\)",
        lambda match: (
            f"![{match.group(1)}]({_get_data_url(match.group(2), image_root)})"
        ),
        markdown,
    )
    return markdown


class GuiProgressBarHandle(_GuiInputHandle[float], GuiProgressBarProps):
    """Handle for updating and removing progress bars."""


class GuiMarkdownHandle(_GuiHandle[None], GuiMarkdownProps):
    """Handling for updating and removing markdown elements."""

    def __init__(self, _impl: _GuiHandleState, _content: str, _image_root: Path | None):
        super().__init__(impl=_impl)
        self._content = _content
        self._image_root = _image_root

    @property
    def content(self) -> str:
        """Current content of this markdown element. Synchronized automatically when assigned."""
        assert self._content is not None
        return self._content

    @content.setter
    def content(self, content: str) -> None:
        self._content = content
        self._markdown = _parse_markdown(content, self._image_root)


class GuiHtmlHandle(_GuiHandle[None], GuiHtmlProps):
    """Handling for updating and removing HTML elements."""


class GuiPlotlyHandle(_GuiHandle[None], GuiPlotlyProps):
    """Handle for updating and removing Plotly figures."""

    def __init__(self, _impl: _GuiHandleState, _figure: go.Figure):
        super().__init__(impl=_impl)
        self._figure = _figure

    @property
    def figure(self) -> go.Figure:
        """Current content of this markdown element. Synchronized automatically when assigned."""
        assert self._figure is not None
        return self._figure

    @figure.setter
    def figure(self, figure: go.Figure) -> None:
        self._figure = figure

        json_str = figure.to_json()
        assert isinstance(json_str, str)
        self._plotly_json_str = json_str


class GuiUplotHandle(_GuiHandle[None], GuiUplotProps):
    """Handle for updating and removing Uplot figures."""

    pass


class GuiImageHandle(_GuiHandle[None], GuiImageProps):
    """Handle for updating and removing images."""

    def __init__(
        self,
        _impl: _GuiHandleState,
        _image: np.ndarray,
        _jpeg_quality: int | None,
    ):
        super().__init__(impl=_impl)
        self._image = _image
        self._jpeg_quality = _jpeg_quality
        self._user_format: Literal["auto", "jpeg", "png"] = "auto"  # Default if not set

    @property
    def image(self) -> np.ndarray:
        """Current content of this image element. Synchronized automatically when assigned."""
        assert self._image is not None
        return self._image

    @image.setter
    def image(self, image: np.ndarray) -> None:
        self._image = image
        resolved_format, data = _encode_image_binary(
            image, self._user_format, jpeg_quality=self._jpeg_quality
        )
        self._format = resolved_format
        self._data = data

    @property
    def format(self) -> Literal["auto", "jpeg", "png"]:
        """Image format. 'auto' will use PNG for RGBA images and JPEG for RGB."""
        return self._user_format

    @format.setter
    def format(self, value: Literal["auto", "jpeg", "png"]) -> None:
        import warnings

        # Skip if format isn't changing.
        if self._user_format == value:
            return

        self._user_format = value

        # Re-encode image.
        if value == "jpeg" and self._image.shape[2] == 4:
            warnings.warn(
                "Converting RGBA image to JPEG will discard the alpha channel."
            )
        resolved_format, data = _encode_image_binary(
            self._image, value, jpeg_quality=self._jpeg_quality
        )
        self._format = resolved_format
        self._data = data



import argparse
import os
import shutil
import subprocess
import sys
from pathlib import Path

client_dir = Path(__file__).absolute().parent / "client"
build_dir = client_dir / "build"


def _is_editable_install() -> bool:
    """Check if viser is installed in editable mode.

    In editable installs, __file__ is in src/viser/ within the source tree.
    In regular installs, __file__ is in site-packages/viser/.
    """
    viser_dir = Path(__file__).parent
    return viser_dir.name == "viser" and viser_dir.parent.name == "src"


def _check_viser_dev_running() -> bool:
    """Returns True if the viewer client has been launched via `npm run dev`."""
    try:
        import psutil
    except ImportError:
        # If psutil is not installed, we can't check for dev server.
        # This is fine for normal usage - only needed for development.
        return False

    for process in psutil.process_iter():
        try:
            # Check if the process is running from the correct viser client directory
            # and is actually a vite dev server (not just any vite command)
            cwd = Path(process.cwd()).resolve()
            expected_cwd = client_dir.resolve()

            if cwd == expected_cwd:
                cmdline = process.cmdline()
                # Check for vite with --host flag (which is our dev command)
                # Make sure it's not a build command
                has_vite = any("vite" in part for part in cmdline)
                has_host = any("--host" in part for part in cmdline)
                not_build = not any("build" in part for part in cmdline)

                if has_vite and has_host and not_build:
                    return True
        except (psutil.AccessDenied, psutil.ZombieProcess, psutil.NoSuchProcess):
            pass
    return False


def ensure_client_is_built() -> None:
    """Ensure that the client is built or already running."""

    # For non-editable installs, just verify the build exists.
    # Skip timestamp checks and dev server detection.
    if not _is_editable_install() and (build_dir / "index.html").exists():
        return

    if not (client_dir / "src").exists():
        # Can't build client.
        assert (build_dir / "index.html").exists(), (
            "Something went wrong! At least one of the client source or build"
            " directories should be present."
        )
        return

    # Do we need to re-trigger a build?
    build = False
    if _check_viser_dev_running():
        # Don't run build if dev server is already running.
        import rich

        rich.print(
            "[bold](viser)[/bold] The Viser viewer looks like it has been launched via"
            " `npm run dev`. Skipping build check..."
        )
        build = False
    elif not (build_dir / "index.html").exists():
        import rich

        rich.print("[bold](viser)[/bold] No client build found. Building now...")
        build = True
    elif (
        # We should be at least 10 seconds newer than the last build.
        # This buffer is important when we install from pip, and the src/ +
        # build/ directories have very similar timestamps.
        _modified_time_recursive(client_dir / "src")
        > _modified_time_recursive(build_dir) + 10.0
    ):
        import rich

        rich.print(
            "[bold](viser)[/bold] Client build looks out of date. Building now..."
        )
        build = True

    # Install nodejs and build if necessary. We assume bash is installed.
    if build:
        _build_viser_client(out_dir=build_dir, cached=False)


def _build_viser_client(out_dir: Path, cached: bool = True) -> None:
    """Create a build of the Viser client.

    Args:
        out_dir: The directory to write the built client to.
        cached: If True, skip the build if the client is already built.
            Instead, we'll simply copy the previous build to the new location.
    """

    if cached and build_dir.exists() and (build_dir / "index.html").exists():
        import rich

        rich.print(
            f"[bold](viser)[/bold] Copying client build from {build_dir} to {out_dir}."
        )
        shutil.copytree(build_dir, out_dir)
        return

    node_bin_dir = _install_sandboxed_node()
    npx_path = node_bin_dir / "npx"

    subprocess_env = os.environ.copy()
    subprocess_env["NODE_VIRTUAL_ENV"] = str(node_bin_dir.parent)
    subprocess_env["PATH"] = (
        str(node_bin_dir)
        + (";" if sys.platform == "win32" else ":")
        + subprocess_env["PATH"]
    )
    npm_path = node_bin_dir / "npm"

    if sys.platform == "win32":
        npx_path = npx_path.with_suffix(".cmd")
        npm_path = npm_path.with_suffix(".cmd")

    subprocess.run(
        args=[str(npm_path), "install"],
        env=subprocess_env,
        cwd=client_dir,
        check=False,
    )
    subprocess.run(
        args=[
            str(npx_path),
            "vite",
            "build",
            "--base",
            "./",
            "--outDir",
            # Relative path needs to be made absolute, since we change the CWD.
            str(out_dir.absolute()),
        ],
        env=subprocess_env,
        cwd=client_dir,
        check=False,
    )


def build_client_entrypoint() -> None:
    """Build the Viser client entrypoint, which is used to launch the viewer."""
    parser = argparse.ArgumentParser(description="Build the Viser client.")
    parser.add_argument("--out-dir", required=True)
    parser.add_argument(
        "--no-cached",
        action="store_false",
        help="If set, skip the build if the client is already built.",
    )
    args = parser.parse_args()
    out_dir = Path(args.out_dir) if args.out_dir else build_dir
    _build_viser_client(out_dir=out_dir, cached=args.no_cached)


def _install_sandboxed_node() -> Path:
    """Install a sandboxed copy of nodejs using nodeenv, and return a path to the
    environment's bin directory (`.nodeenv/bin` or `.nodeenv/Scripts`).

    On Windows, the `.nodeenv/bin` does not exist. Instead, executables are
    installed to `.nodeenv/Scripts`."""

    def get_node_bin_dir() -> Path:
        env_dir = client_dir / ".nodeenv"
        node_bin_dir = env_dir / "bin"
        if not node_bin_dir.exists():
            node_bin_dir = env_dir / "Scripts"
        return node_bin_dir

    node_bin_dir = get_node_bin_dir()
    if (node_bin_dir / "npx").exists():
        import rich

        rich.print("[bold](viser)[/bold] nodejs is set up!")
        return node_bin_dir

    env_dir = client_dir / ".nodeenv"
    result = subprocess.run(
        [sys.executable, "-m", "nodeenv", "--node=24.12.0", env_dir], check=False
    )

    if result.returncode != 0:
        raise RuntimeError(
            "Failed to install Node.js using nodeenv. "
            "To rebuild the Viser client, install nodeenv with: "
            "pip install 'nodeenv>=1.9.1'"
        )

    node_bin_dir = get_node_bin_dir()
    assert (node_bin_dir / "npx").exists()
    return node_bin_dir


def _modified_time_recursive(dir: Path) -> float:
    """Recursively get the last time a file was modified in a directory."""
    return max([f.stat().st_mtime for f in dir.glob("**/*")])



from __future__ import annotations

import sys
import traceback
from concurrent.futures import Future
from typing import Any


def print_threadpool_errors(future: Future[Any]) -> None:
    """Print errors from a Future in a ThreadPool, should be used with
    `add_done_callback`."""
    if future.cancelled():
        print("Task was cancelled", file=sys.stderr)
        return

    exc = future.exception()
    if exc is not None:
        print("Task failed with exception:", file=sys.stderr)
        traceback.print_exception(type(exc), exc, exc.__traceback__)



from __future__ import annotations

import copy
import dataclasses
import warnings
from collections.abc import Coroutine
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Literal,
    Protocol,
    TypeVar,
    Union,
    cast,
)

import numpy as np
import numpy.typing as npt
from typing_extensions import Self, deprecated, override

from . import _messages
from ._assignable_props_api import AssignablePropsBase
from .infra._infra import WebsockClientConnection, WebsockServer

if TYPE_CHECKING:
    from ._gui_api import GuiApi
    from ._scene_api import SceneApi
    from ._viser import ClientHandle
    from .infra import ClientId


@dataclasses.dataclass(frozen=True)
class ScenePointerEvent:
    """Event passed to pointer callbacks for the scene (currently only clicks)."""

    client: ClientHandle
    """Client that triggered this event."""
    client_id: int
    """ID of client that triggered this event."""
    event_type: _messages.ScenePointerEventType
    """Type of event that was triggered. Currently we only support clicks and box selections."""
    ray_origin: tuple[float, float, float] | None
    """Origin of 3D ray corresponding to this click, in world coordinates."""
    ray_direction: tuple[float, float, float] | None
    """Direction of 3D ray corresponding to this click, in world coordinates."""
    screen_pos: tuple[tuple[float, float], ...]
    """Screen position of the click on the screen (OpenCV image coordinates, 0 to 1).
    (0, 0) is the upper-left corner, (1, 1) is the bottom-right corner.
    For a box selection, this includes the min- and max- corners of the box."""

    @property
    @deprecated("The `event` property is deprecated. Use `event_type` instead.")
    def event(self):
        """Deprecated. Use `event_type` instead.

        .. deprecated:: 0.2.23
            The `event` property is deprecated. Use `event_type` instead.
        """
        return self.event_type


TSceneNodeHandle = TypeVar("TSceneNodeHandle", bound="SceneNodeHandle")


@dataclasses.dataclass
class _SceneNodeHandleState:
    name: str
    props: Any  # _messages.*Prop object.
    """Message containing properties of this scene node that are sent to the
    client."""
    api: SceneApi
    wxyz: np.ndarray = dataclasses.field(
        default_factory=lambda: np.array([1.0, 0.0, 0.0, 0.0])
    )
    position: np.ndarray = dataclasses.field(
        default_factory=lambda: np.array([0.0, 0.0, 0.0])
    )
    visible: bool = True
    click_cb: list[
        Callable[[SceneNodePointerEvent[_ClickableSceneNodeHandle]], None | Coroutine]
    ] = dataclasses.field(default_factory=list)
    removed: bool = False


class _SceneNodeMessage(Protocol):
    name: str
    props: Any


class SceneNodeHandle(AssignablePropsBase[_SceneNodeHandleState]):
    """Handle base class for interacting with scene nodes."""

    @override
    def _queue_update(self, name: str, value: Any) -> None:
        self._impl.api._websock_interface.queue_message(
            _messages.SceneNodeUpdateMessage(self._impl.name, {name: value})
        )

    @property
    def name(self) -> str:
        """Read-only name of the scene node."""
        return self._impl.name

    @classmethod
    def _make(
        cls: type[TSceneNodeHandle],
        api: SceneApi,
        message: _SceneNodeMessage,
        name: str,
        wxyz: tuple[float, float, float, float] | np.ndarray,
        position: tuple[float, float, float] | np.ndarray,
        visible: bool,
    ) -> TSceneNodeHandle:
        """Create scene node: send state to client(s) and set up
        server-side state."""
        # Normalize name to always start with "/".
        if not name.startswith("/"):
            name = "/" + name
            message.name = name

        # Ensure all ancestor nodes exist (creates intermediate frames as needed).
        api._ensure_ancestors_exist(name)

        # Send message.
        assert isinstance(message, _messages.Message)
        api._websock_interface.queue_message(message)

        out = cls(_SceneNodeHandleState(name, copy.deepcopy(message.props), api))
        api._handle_from_node_name[name] = out

        # Track parent -> child relationship.
        parent = name.rsplit("/", 1)[0]
        api._children_from_node_name.setdefault(parent, set()).add(name)
        api._children_from_node_name.setdefault(name, set())

        out.wxyz = wxyz
        out.position = position

        # Toggle visibility to make sure we send a
        # SetSceneNodeVisibilityMessage to the client.
        out._impl.visible = not visible
        out.visible = visible
        return out

    @property
    def wxyz(self) -> npt.NDArray[np.float64]:
        """Orientation of the scene node. This is the quaternion representation of the R
        in `p_parent = [R | t] p_local`. Synchronized to clients automatically when assigned.
        """
        return self._impl.wxyz

    @wxyz.setter
    def wxyz(self, wxyz: tuple[float, float, float, float] | np.ndarray) -> None:
        from ._scene_api import cast_vector

        wxyz_cast = cast_vector(wxyz, 4)
        wxyz_array = np.asarray(wxyz)
        if np.allclose(wxyz_cast, self._impl.wxyz):
            return
        self._impl.wxyz[:] = wxyz_array
        self._impl.api._websock_interface.queue_message(
            _messages.SetOrientationMessage(self._impl.name, wxyz_cast)
        )

    @property
    def position(self) -> npt.NDArray[np.float64]:
        """Position of the scene node. This is equivalent to the t in
        `p_parent = [R | t] p_local`. Synchronized to clients automatically when assigned.
        """
        return self._impl.position

    @position.setter
    def position(self, position: tuple[float, float, float] | np.ndarray) -> None:
        from ._scene_api import cast_vector

        position_cast = cast_vector(position, 3)
        position_array = np.asarray(position)
        if np.allclose(position_array, self._impl.position):
            return
        self._impl.position[:] = position_array
        self._impl.api._websock_interface.queue_message(
            _messages.SetPositionMessage(self._impl.name, position_cast)
        )

    @property
    def visible(self) -> bool:
        """Whether the scene node is visible or not. Synchronized to clients automatically when assigned."""
        return self._impl.visible

    @visible.setter
    def visible(self, visible: bool) -> None:
        if visible == self._impl.visible:
            return
        self._impl.api._websock_interface.queue_message(
            _messages.SetSceneNodeVisibilityMessage(self._impl.name, visible)
        )
        self._impl.visible = visible

    def remove(self) -> None:
        """Remove the node from the scene."""
        # Warn if already removed.
        if self._impl.removed:
            warnings.warn(f"Attempted to remove already removed node: {self.name}")
            return

        # Collect all descendants via BFS.
        to_remove = [self._impl.name]
        i = 0
        while i < len(to_remove):
            children = self._impl.api._children_from_node_name.get(to_remove[i], ())
            to_remove.extend(children)
            i += 1

        # Clean up all descendants from both dicts.
        for node_name in to_remove:
            handle = self._impl.api._handle_from_node_name.pop(node_name, None)
            if handle is not None:
                handle._impl.removed = True
            self._impl.api._children_from_node_name.pop(node_name, None)

        # Remove from parent's children set.
        parent = self._impl.name.rsplit("/", 1)[0]
        parent_children = self._impl.api._children_from_node_name.get(parent)
        if parent_children is not None:
            parent_children.discard(self._impl.name)

        # Single message — client cascades removal to children.
        self._impl.api._websock_interface.queue_message(
            _messages.RemoveSceneNodeMessage(self._impl.name)
        )


@dataclasses.dataclass(frozen=True)
class SceneNodePointerEvent(Generic[TSceneNodeHandle]):
    """Event passed to pointer callbacks for scene nodes (currently only clicks)."""

    client: ClientHandle
    """Client that triggered this event."""
    client_id: int
    """ID of client that triggered this event."""
    event: Literal["click"]
    """Type of event that was triggered. Currently we only support clicks."""
    target: TSceneNodeHandle
    """Scene node that was clicked."""
    ray_origin: tuple[float, float, float]
    """Origin of 3D ray corresponding to this click, in world coordinates."""
    ray_direction: tuple[float, float, float]
    """Direction of 3D ray corresponding to this click, in world coordinates."""
    screen_pos: tuple[float, float]
    """Screen position of the click on the screen (OpenCV image coordinates, 0 to 1).
    (0, 0) is the upper-left corner, (1, 1) is the bottom-right corner."""
    instance_index: int | None
    """Instance ID of the clicked object, if applicable. Currently this is `None` for all objects except for the output of :meth:`SceneApi.add_batched_axes()`."""


@dataclasses.dataclass(frozen=True)
class TransformControlsEvent:
    """Event passed to callbacks for transform control updates."""

    client: ClientHandle | None
    """Client that triggered this event."""
    client_id: int | None
    """ID of client that triggered this event."""
    target: TransformControlsHandle
    """Transform controls handle that was affected."""


NoneOrCoroutine = TypeVar("NoneOrCoroutine", None, Coroutine)


class _ClickableSceneNodeHandle(SceneNodeHandle):
    def on_click(
        self: Self,
        func: Callable[[SceneNodePointerEvent[Self]], NoneOrCoroutine],
    ) -> Callable[[SceneNodePointerEvent[Self]], NoneOrCoroutine]:
        """Attach a callback for when a scene node is clicked.

        The callback can be either a standard function or an async function:
        - Standard functions (def) will be executed in a threadpool.
        - Async functions (async def) will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._impl.api._websock_interface.queue_message(
            _messages.SetSceneNodeClickableMessage(self._impl.name, True)
        )
        if self._impl.click_cb is None:
            self._impl.click_cb = []
        self._impl.click_cb.append(
            cast(
                Callable[
                    [SceneNodePointerEvent[_ClickableSceneNodeHandle]],
                    # `Union[X, Y]` instead of `X | Y` for Python 3.8 support.
                    Union[None, Coroutine],
                ],
                func,
            )
        )
        return func

    def remove_click_callback(
        self, callback: Literal["all"] | Callable = "all"
    ) -> None:
        """Remove click callbacks from scene node.

        Args:
            callback: Either "all" to remove all callbacks, or a specific callback function to remove.
        """
        if callback == "all":
            self._impl.click_cb.clear()
        else:
            self._impl.click_cb = [cb for cb in self._impl.click_cb if cb != callback]
        if len(self._impl.click_cb) == 0:
            self._impl.api._websock_interface.queue_message(
                _messages.SetSceneNodeClickableMessage(self._impl.name, False)
            )

    def remove(self) -> None:
        """Remove the node from the scene."""
        if len(self._impl.click_cb) > 0:
            # SetSceneNodeClickableMessage can still remain in the message
            # buffer, even if a scene node is removed. This could cause a new
            # scene node to be mistakenly considered clickable if it is made
            # with the same name. We ideally would fix this with better state
            # management... in the meantime we can just set the flag to False.
            self._impl.api._websock_interface.queue_message(
                _messages.SetSceneNodeClickableMessage(self._impl.name, False)
            )
        super().remove()


class CameraFrustumHandle(
    _ClickableSceneNodeHandle,
    _messages.CameraFrustumProps,
):
    """Handle for camera frustums."""

    _image: np.ndarray | None
    _jpeg_quality: int | None
    _user_format: Literal["auto", "jpeg", "png"]

    @property
    def image(self) -> np.ndarray | None:
        """Current content of the image. Synchronized automatically when assigned."""
        return self._image

    @image.setter
    def image(self, image: np.ndarray | None) -> None:
        from ._scene_api import _encode_image_binary

        if image is None:
            self._image_data = None
            return

        self._image = image
        resolved_format, data = _encode_image_binary(
            image, self._user_format, jpeg_quality=self._jpeg_quality
        )
        self._format = resolved_format
        self._image_data = data

    @property
    def format(self) -> Literal["auto", "jpeg", "png"]:
        """Image format. 'auto' will use PNG for RGBA images and JPEG for RGB."""
        return self._user_format

    @format.setter
    def format(self, value: Literal["auto", "jpeg", "png"]) -> None:
        import warnings

        from ._scene_api import _encode_image_binary

        # Skip if format isn't changing.
        if self._user_format == value:
            return

        self._user_format = value

        # Re-encode image. if we have one.
        if self._image is not None:
            if value == "jpeg" and self._image.shape[2] == 4:
                warnings.warn(
                    "Converting RGBA image to JPEG will discard the alpha channel."
                )
            resolved_format, data = _encode_image_binary(
                self._image, value, jpeg_quality=self._jpeg_quality
            )
            self._format = resolved_format
            self._image_data = data

    def compute_canonical_frustum_size(self) -> tuple[float, float, float]:
        """Compute the X, Y, and Z dimensions of the frustum if it had
        `.scale=1.0`. These dimensions will change whenever `.fov` or `.aspect`
        are changed.

        To set the distance between a frustum's origin and image plane to 1, we
        can run:

        .. code-block:: python

            frustum.scale = 1.0 / frustum.compute_canonical_frustum_size()[2]


        `.scale` can be a float for uniform scaling or a 3-tuple for per-axis
        scaling of the X, Y, and Z dimensions. It aims to preserve the visual
        volume of the frustum regardless of the aspect ratio or FOV. This
        method allows more precise computation and control of the frustum's
        dimensions.
        """
        # Math used in the client implementation.
        y = np.tan(self.fov / 2.0)
        x = y * self.aspect
        z = 1.0
        volume_scale = np.cbrt((x * y * z) / 3.0)

        z /= volume_scale

        # x and y need to be doubled, since on the client they correspond to
        # NDC-style spans [-1, 1].
        return x * 2.0, y * 2.0, z


class DirectionalLightHandle(
    SceneNodeHandle,
    _messages.DirectionalLightProps,
):
    """Handle for directional lights."""


class AmbientLightHandle(
    SceneNodeHandle,
    _messages.AmbientLightProps,
):
    """Handle for ambient lights."""


class HemisphereLightHandle(
    SceneNodeHandle,
    _messages.HemisphereLightProps,
):
    """Handle for hemisphere lights."""


class PointLightHandle(
    SceneNodeHandle,
    _messages.PointLightProps,
):
    """Handle for point lights."""


class RectAreaLightHandle(
    SceneNodeHandle,
    _messages.RectAreaLightProps,
):
    """Handle for rectangular area lights."""


class SpotLightHandle(
    SceneNodeHandle,
    _messages.SpotLightProps,
):
    """Handle for spot lights."""


class PointCloudHandle(
    SceneNodeHandle,
    _messages.PointCloudProps,
):
    """Handle for point clouds. Does not support click events."""

    @override
    def _cast_array_dtypes(
        self,
        prop_hints: Dict[str, Any],
        prop_name: str,
        value: np.ndarray,
    ) -> np.ndarray:
        """Casts assigned `points` based on the current value of `precision`."""
        if prop_name == "points":
            return value.astype(
                {"float16": np.float16, "float32": np.float32}[self.precision]
            )
        return super()._cast_array_dtypes(prop_hints, prop_name, value)


class BatchedAxesHandle(
    _ClickableSceneNodeHandle,
    _messages.BatchedAxesProps,
):
    """Handle for batched coordinate frames."""


class FrameHandle(
    _ClickableSceneNodeHandle,
    _messages.FrameProps,
):
    """Handle for coordinate frames."""


class MeshHandle(
    _ClickableSceneNodeHandle,
    _messages.MeshProps,
):
    """Handle for mesh objects."""


class BoxHandle(
    _ClickableSceneNodeHandle,
    _messages.BoxProps,
):
    """Handle for box objects."""


class IcosphereHandle(
    _ClickableSceneNodeHandle,
    _messages.IcosphereProps,
):
    """Handle for icosphere objects."""


class CylinderHandle(
    _ClickableSceneNodeHandle,
    _messages.CylinderProps,
):
    """Handle for cylinder objects."""


class BatchedMeshHandle(
    _ClickableSceneNodeHandle,
    _messages.BatchedMeshesProps,
):
    """Handle for batched mesh objects."""


class BatchedGlbHandle(
    _ClickableSceneNodeHandle,
    _messages.BatchedGlbProps,
):
    """Handle for batched GLB objects."""


class GaussianSplatHandle(
    _ClickableSceneNodeHandle,
    _messages.GaussianSplatsProps,
):
    """Handle for Gaussian splatting objects.

    **Work-in-progress.** Gaussian rendering is still under development.

    Buffer layout per Gaussian (8 uint32 elements = 32 bytes):
        - [0:3]: centers (3x float32)
        - [3]: reserved for renderer
        - [4:7]: covariance upper-triangular (6x float16)
        - [7]: RGBA (4x uint8)
    """

    def _ensure_buffer_size(self, num_gaussians: int) -> None:
        """Ensure the internal buffer can hold the specified number of Gaussians.

        If the buffer is already the correct size, this is a no-op. Otherwise,
        a new buffer is allocated with default values (white color, full opacity,
        small identity-like covariances, centers at origin).
        """
        if self.buffer.shape[0] == num_gaussians:
            return

        # Create new buffer with default values.
        new_buffer = np.zeros((num_gaussians, 8), dtype=np.uint32)

        # Set default RGBA to white, fully opaque (255, 255, 255, 255).
        new_buffer[:, 7] = 0xFFFFFFFF

        # Set default covariances to small identity-like values.
        # Store as 6 float16 values: [cov00, cov01, cov02, cov11, cov12, cov22].
        default_cov = np.array([0.01, 0.0, 0.0, 0.01, 0.0, 0.01], dtype=np.float16)
        new_buffer[:, 4:7] = np.tile(default_cov.view(np.uint32), (num_gaussians, 1))

        self.buffer = new_buffer

    @property
    def centers(self) -> npt.NDArray[np.float32]:
        """Centers of the Gaussians. Shape: (N, 3). Synchronized automatically when assigned."""
        return self.buffer[:, 0:3].view(np.float32)

    @centers.setter
    def centers(self, centers: np.ndarray) -> None:
        assert centers.ndim == 2 and centers.shape[1] == 3, (
            f"centers must have shape (N, 3), got {centers.shape}"
        )
        self._ensure_buffer_size(centers.shape[0])
        self.buffer[:, 0:3] = centers.astype(np.float32).view(np.uint32)
        self._queue_update("buffer", self.buffer)

    @property
    def rgbs(self) -> npt.NDArray[np.uint8]:
        """Colors of the Gaussians. Shape: (N, 3). Values in [0, 1]. Synchronized automatically when assigned."""
        rgba = self.buffer[:, 7:8].view(np.uint8).reshape(-1, 4)
        return rgba[:, :3]

    @rgbs.setter
    def rgbs(self, rgbs: np.ndarray) -> None:
        from ._assignable_props_api import colors_to_uint8

        assert rgbs.ndim == 2 and rgbs.shape[1] == 3, (
            f"rgbs must have shape (N, 3), got {rgbs.shape}"
        )
        self._ensure_buffer_size(rgbs.shape[0])
        rgba = self.buffer[:, 7:8].view(np.uint8).reshape(-1, 4)
        rgba[:, :3] = colors_to_uint8(rgbs)
        self.buffer[:, 7:8] = rgba.view(np.uint32)
        self._queue_update("buffer", self.buffer)

    @property
    def opacities(self) -> npt.NDArray[np.uint8]:
        """Opacities of the Gaussians. Shape: (N, 1). Values in [0, 1]. Synchronized automatically when assigned."""
        buffer = self.buffer
        rgba = buffer[:, 7:8].view(np.uint8).reshape(-1, 4)
        return rgba[:, 3:4]

    @opacities.setter
    def opacities(self, opacities: np.ndarray) -> None:
        from ._assignable_props_api import colors_to_uint8

        assert opacities.ndim == 2 and opacities.shape[1] == 1, (
            f"opacities must have shape (N, 1), got {opacities.shape}"
        )
        self._ensure_buffer_size(opacities.shape[0])
        rgba = self.buffer[:, 7:8].view(np.uint8).reshape(-1, 4)
        rgba[:, 3:4] = colors_to_uint8(opacities)
        self.buffer[:, 7:8] = rgba.view(np.uint32)
        self._queue_update("buffer", self.buffer)

    @property
    def covariances(self) -> npt.NDArray[np.float32]:
        """Covariances of the Gaussians. Shape: (N, 3, 3). Synchronized automatically when assigned."""
        # Extract upper-triangular terms stored as 6 float16 values.
        cov_triu_f16 = self.buffer[:, 4:7].view(np.float16).reshape(-1, 6)
        cov_triu = cov_triu_f16.astype(np.float32)
        # Reconstruct symmetric 3x3 matrix.
        n = cov_triu.shape[0]
        cov = np.zeros((n, 3, 3), dtype=np.float32)
        cov[:, 0, 0] = cov_triu[:, 0]
        cov[:, 0, 1] = cov_triu[:, 1]
        cov[:, 0, 2] = cov_triu[:, 2]
        cov[:, 1, 0] = cov_triu[:, 1]  # Symmetric.
        cov[:, 1, 1] = cov_triu[:, 3]
        cov[:, 1, 2] = cov_triu[:, 4]
        cov[:, 2, 0] = cov_triu[:, 2]  # Symmetric.
        cov[:, 2, 1] = cov_triu[:, 4]  # Symmetric.
        cov[:, 2, 2] = cov_triu[:, 5]
        return cov

    @covariances.setter
    def covariances(self, covariances: np.ndarray) -> None:
        assert covariances.ndim == 3 and covariances.shape[1:] == (3, 3), (
            f"covariances must have shape (N, 3, 3), got {covariances.shape}"
        )
        self._ensure_buffer_size(covariances.shape[0])
        # Extract upper-triangular terms: indices [0,1,2,4,5,8] from flattened 3x3.
        cov_triu = covariances.reshape((-1, 9))[:, np.array([0, 1, 2, 4, 5, 8])]
        cov_triu_f16 = cov_triu.astype(np.float16)
        self.buffer[:, 4:7] = np.ascontiguousarray(cov_triu_f16).view(np.uint32)
        self._queue_update("buffer", self.buffer)


class MeshSkinnedHandle(
    _ClickableSceneNodeHandle,
    _messages.SkinnedMeshProps,
):
    """Handle for skinned mesh objects."""

    def __init__(
        self, impl: _SceneNodeHandleState, bones: tuple[MeshSkinnedBoneHandle, ...]
    ):
        super().__init__(impl)
        self.bones = bones


@dataclasses.dataclass
class BoneState:
    name: str
    websock_interface: WebsockServer | WebsockClientConnection
    bone_index: int
    wxyz: np.ndarray
    position: np.ndarray


@dataclasses.dataclass
class MeshSkinnedBoneHandle:
    """Handle for reading and writing the poses of bones in a skinned mesh."""

    _impl: BoneState

    @property
    def wxyz(self) -> npt.NDArray[np.float64]:
        """Orientation of the bone. This is the quaternion representation of the R
        in `p_parent = [R | t] p_local`. Synchronized to clients automatically when assigned.
        """
        return self._impl.wxyz

    @wxyz.setter
    def wxyz(self, wxyz: tuple[float, float, float, float] | np.ndarray) -> None:
        from ._scene_api import cast_vector

        wxyz_cast = cast_vector(wxyz, 4)
        wxyz_array = np.asarray(wxyz)
        if np.allclose(wxyz_cast, self._impl.wxyz):
            return
        self._impl.wxyz[:] = wxyz_array
        self._impl.websock_interface.queue_message(
            _messages.SetBoneOrientationMessage(
                self._impl.name, self._impl.bone_index, wxyz_cast
            )
        )

    @property
    def position(self) -> npt.NDArray[np.float64]:
        """Position of the bone. This is equivalent to the t in
        `p_parent = [R | t] p_local`. Synchronized to clients automatically when assigned.
        """
        return self._impl.position

    @position.setter
    def position(self, position: tuple[float, float, float] | np.ndarray) -> None:
        from ._scene_api import cast_vector

        position_cast = cast_vector(position, 3)
        position_array = np.asarray(position)
        if np.allclose(position_array, self._impl.position):
            return
        self._impl.position[:] = position_array
        self._impl.websock_interface.queue_message(
            _messages.SetBonePositionMessage(
                self._impl.name, self._impl.bone_index, position_cast
            )
        )


class GridHandle(
    SceneNodeHandle,
    _messages.GridProps,
):
    """Handle for grid objects."""


class LineSegmentsHandle(
    SceneNodeHandle,
    _messages.LineSegmentsProps,
):
    """Handle for line segments objects."""


class SplineCatmullRomHandle(
    SceneNodeHandle,
    _messages.CatmullRomSplineProps,
):
    """Handle for Catmull-Rom splines."""

    @property
    @deprecated("The 'positions' property is deprecated. Use 'points' instead.")
    def positions(self) -> tuple[tuple[float, float, float], ...]:
        """Get the spline positions. Deprecated: use 'points' instead.

        .. deprecated:: 1.0.0
            "The 'positions' tuple property is deprecated. Use the 'points' numpy array instead.",
        """
        import warnings

        warnings.warn(
            "The 'positions' tuple property is deprecated. Use the 'points' numpy array instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return tuple(tuple(x) for x in self.points.tolist())  # type: ignore

    @positions.setter
    @deprecated("The 'positions' property is deprecated. Use 'points' instead.")
    def positions(self, positions: tuple[tuple[float, float, float], ...]) -> None:
        import warnings

        warnings.warn(
            "The 'positions' tuple property is deprecated. Use the 'points' numpy array instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.points = np.asarray(positions)


class SplineCubicBezierHandle(
    SceneNodeHandle,
    _messages.CubicBezierSplineProps,
):
    """Handle for cubic Bezier splines."""

    @property
    @deprecated(
        "The 'positions' tuple property is deprecated. Use 'points' numpy array instead."
    )
    def positions(self) -> tuple[tuple[float, float, float], ...]:
        """Get the spline positions. Deprecated: use 'points' instead.

        .. deprecated:: 1.0.0
            The 'positions' tuple property is deprecated. Use the 'points' numpy array instead.
        """
        return tuple(tuple(p) for p in self.points.tolist())  # type: ignore

    @positions.setter
    @deprecated(
        "The 'positions' tuple property is deprecated. Use the 'points' numpy array instead."
    )
    def positions(self, positions: tuple[tuple[float, float, float], ...]) -> None:
        import warnings

        warnings.warn(
            "The 'positions' tuple property is deprecated. Use the 'points' numpy array instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.points = np.asarray(positions)


class GlbHandle(
    _ClickableSceneNodeHandle,
    _messages.GlbProps,
):
    """Handle for GLB objects."""


class ImageHandle(
    _ClickableSceneNodeHandle,
    _messages.ImageProps,
):
    """Handle for 2D images, rendered in 3D."""

    _image: np.ndarray
    _jpeg_quality: int | None
    _user_format: Literal["auto", "jpeg", "png"]

    @property
    def image(self) -> np.ndarray:
        """Current content of the image. Synchronized automatically when assigned."""
        assert self._image is not None
        return self._image

    @image.setter
    def image(self, image: np.ndarray) -> None:
        from ._scene_api import _encode_image_binary

        self._image = image
        resolved_format, data = _encode_image_binary(
            image, self._user_format, jpeg_quality=self._jpeg_quality
        )
        self._format = resolved_format
        self._data = data

    @property
    def format(self) -> Literal["auto", "jpeg", "png"]:
        """Image format. 'auto' will use PNG for RGBA images and JPEG for RGB."""
        return self._user_format

    @format.setter
    def format(self, value: Literal["auto", "jpeg", "png"]) -> None:
        import warnings

        from ._scene_api import _encode_image_binary

        # Skip if format isn't changing.
        if self._user_format == value:
            return

        self._user_format = value

        # Re-encode image.
        if value == "jpeg" and self._image.shape[2] == 4:
            warnings.warn(
                "Converting RGBA image to JPEG will discard the alpha channel."
            )
        resolved_format, data = _encode_image_binary(
            self._image, value, jpeg_quality=self._jpeg_quality
        )
        self._format = resolved_format
        self._data = data


class LabelHandle(
    SceneNodeHandle,
    _messages.LabelProps,
):
    """Handle for 2D label objects. Does not support click events."""


@dataclasses.dataclass
class _TransformControlsState:
    last_updated: float
    update_cb: list[Callable[[TransformControlsEvent], None | Coroutine]]
    drag_start_cb: list[Callable[[TransformControlsEvent], None | Coroutine]] = (
        dataclasses.field(default_factory=list)
    )
    drag_end_cb: list[Callable[[TransformControlsEvent], None | Coroutine]] = (
        dataclasses.field(default_factory=list)
    )
    sync_cb: None | Callable[[ClientId, TransformControlsHandle], None] = None


class TransformControlsHandle(
    SceneNodeHandle,
    _messages.TransformControlsProps,
):
    """Handle for interacting with transform control gizmos."""

    def __init__(self, impl: _SceneNodeHandleState, impl_aux: _TransformControlsState):
        super().__init__(impl)
        self._impl_aux = impl_aux

    @property
    def update_timestamp(self) -> float:
        return self._impl_aux.last_updated

    def on_update(
        self, func: Callable[[TransformControlsEvent], NoneOrCoroutine]
    ) -> Callable[[TransformControlsEvent], NoneOrCoroutine]:
        """Attach a callback for when the gizmo is moved.

        The callback can be either a standard function or an async function:
        - Standard functions (def) will be executed in a threadpool.
        - Async functions (async def) will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._impl_aux.update_cb.append(func)
        return func

    def remove_update_callback(
        self, callback: Literal["all"] | Callable = "all"
    ) -> None:
        """Remove update callbacks from the transform controls.

        Args:
            callback: Either "all" to remove all callbacks, or a specific callback function to remove.
        """
        if callback == "all":
            self._impl_aux.update_cb.clear()
        else:
            self._impl_aux.update_cb = [
                cb for cb in self._impl_aux.update_cb if cb != callback
            ]

    def on_drag_start(
        self, func: Callable[[TransformControlsEvent], NoneOrCoroutine]
    ) -> Callable[[TransformControlsEvent], NoneOrCoroutine]:
        """Attach a callback for when dragging starts ("mouse down").

        The callback can be either a standard function or an async function:
        - Standard functions (def) will be executed in a threadpool.
        - Async functions (async def) will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._impl_aux.drag_start_cb.append(func)
        return func

    def on_drag_end(
        self, func: Callable[[TransformControlsEvent], NoneOrCoroutine]
    ) -> Callable[[TransformControlsEvent], NoneOrCoroutine]:
        """Attach a callback for when dragging end ("mouse up").

        The callback can be either a standard function or an async function:
        - Standard functions (def) will be executed in a threadpool.
        - Async functions (async def) will be executed in the event loop.

        Using async functions can be useful for reducing race conditions.
        """
        self._impl_aux.drag_end_cb.append(func)
        return func

    def remove_drag_start_callback(
        self, callback: Literal["all"] | Callable = "all"
    ) -> None:
        """Remove drag start callbacks from the transform controls.

        Args:
            callback: Either "all" to remove all callbacks, or a specific callback function to remove.
        """
        if callback == "all":
            self._impl_aux.drag_start_cb.clear()
        else:
            self._impl_aux.drag_start_cb = [
                cb for cb in self._impl_aux.drag_start_cb if cb != callback
            ]

    def remove_drag_end_callback(
        self, callback: Literal["all"] | Callable = "all"
    ) -> None:
        """Remove drag end callbacks from the transform controls.

        Args:
            callback: Either "all" to remove all callbacks, or a specific callback function to remove.
        """
        if callback == "all":
            self._impl_aux.drag_end_cb.clear()
        else:
            self._impl_aux.drag_end_cb = [
                cb for cb in self._impl_aux.drag_end_cb if cb != callback
            ]

    def remove(self) -> None:
        """Remove the node from the scene."""
        self._impl.api._handle_from_transform_controls_name.pop(self.name)
        super().remove()


class Gui3dContainerHandle(
    SceneNodeHandle,
    _messages.Gui3DProps,
):
    """Use as a context to place GUI elements into a 3D GUI container."""

    def __init__(self, impl: _SceneNodeHandleState, gui_api: GuiApi, container_id: str):
        super().__init__(impl)
        self._gui_api = gui_api
        self._container_id = container_id
        self._container_id_restore = None
        self._children = {}
        self._gui_api._container_handle_from_uuid[self._container_id] = self

    def __enter__(self) -> Gui3dContainerHandle:
        self._container_id_restore = self._gui_api._get_container_uuid()
        self._gui_api._set_container_uuid(self._container_id)
        return self

    def __exit__(self, *args) -> None:
        del args
        assert self._container_id_restore is not None
        self._gui_api._set_container_uuid(self._container_id_restore)
        self._container_id_restore = None

    def remove(self) -> None:
        """Permanently remove this GUI container from the visualizer."""

        # Call scene node remove.
        super().remove()

        # Clean up contained GUI elements.
        for child in tuple(self._children.values()):
            child.remove()
        self._gui_api._container_handle_from_uuid.pop(self._container_id)



from typing import Literal, Optional, Tuple, TypedDict


class TitlebarButton(TypedDict):
    """A link-only button that appears in the Titlebar."""

    text: Optional[str]
    icon: Optional[Literal["GitHub", "Description", "Keyboard"]]
    href: Optional[str]


class TitlebarImage(TypedDict):
    """An image that appears on the titlebar."""

    image_url_light: str
    image_url_dark: Optional[str]
    image_alt: str
    href: Optional[str]


class TitlebarConfig(TypedDict):
    """Configure the content that appears in the titlebar."""

    buttons: Optional[Tuple[TitlebarButton, ...]]
    image: Optional[TitlebarImage]



from ._gui_api import GuiApi as GuiApi
from ._gui_handles import GuiButtonGroupHandle as GuiButtonGroupHandle
from ._gui_handles import GuiButtonHandle as GuiButtonHandle
from ._gui_handles import GuiCheckboxHandle as GuiCheckboxHandle
from ._gui_handles import GuiDropdownHandle as GuiDropdownHandle
from ._gui_handles import GuiEvent as GuiEvent
from ._gui_handles import GuiFolderHandle as GuiFolderHandle
from ._gui_handles import GuiHtmlHandle as GuiHtmlHandle
from ._gui_handles import GuiImageHandle as GuiImageHandle
from ._gui_handles import GuiInputHandle as GuiInputHandle
from ._gui_handles import GuiMarkdownHandle as GuiMarkdownHandle
from ._gui_handles import GuiMultiSliderHandle as GuiMultiSliderHandle
from ._gui_handles import GuiNumberHandle as GuiNumberHandle
from ._gui_handles import GuiPlotlyHandle as GuiPlotlyHandle
from ._gui_handles import GuiRgbaHandle as GuiRgbaHandle
from ._gui_handles import GuiRgbHandle as GuiRgbHandle
from ._gui_handles import GuiSliderHandle as GuiSliderHandle
from ._gui_handles import GuiTabGroupHandle as GuiTabGroupHandle
from ._gui_handles import GuiTabHandle as GuiTabHandle
from ._gui_handles import GuiTextHandle as GuiTextHandle
from ._gui_handles import GuiUploadButtonHandle as GuiUploadButtonHandle
from ._gui_handles import GuiUplotHandle as GuiUplotHandle
from ._gui_handles import GuiVector2Handle as GuiVector2Handle
from ._gui_handles import GuiVector3Handle as GuiVector3Handle
from ._gui_handles import UploadedFile as UploadedFile
from ._icons_enum import Icon as Icon
from ._icons_enum import IconName as IconName
from ._notification_handle import NotificationHandle as NotificationHandle
from ._scene_api import SceneApi as SceneApi
from ._scene_handles import AmbientLightHandle as AmbientLightHandle
from ._scene_handles import BatchedAxesHandle as BatchedAxesHandle
from ._scene_handles import BatchedGlbHandle as BatchedGlbHandle
from ._scene_handles import BatchedMeshHandle as BatchedMeshHandle
from ._scene_handles import BoxHandle as BoxHandle
from ._scene_handles import CameraFrustumHandle as CameraFrustumHandle
from ._scene_handles import CylinderHandle as CylinderHandle
from ._scene_handles import DirectionalLightHandle as DirectionalLightHandle
from ._scene_handles import FrameHandle as FrameHandle
from ._scene_handles import GaussianSplatHandle as GaussianSplatHandle
from ._scene_handles import GlbHandle as GlbHandle
from ._scene_handles import GridHandle as GridHandle
from ._scene_handles import Gui3dContainerHandle as Gui3dContainerHandle
from ._scene_handles import HemisphereLightHandle as HemisphereLightHandle
from ._scene_handles import IcosphereHandle as IcosphereHandle
from ._scene_handles import ImageHandle as ImageHandle
from ._scene_handles import LabelHandle as LabelHandle
from ._scene_handles import LineSegmentsHandle as LineSegmentsHandle
from ._scene_handles import MeshHandle as MeshHandle
from ._scene_handles import MeshSkinnedBoneHandle as MeshSkinnedBoneHandle
from ._scene_handles import MeshSkinnedHandle as MeshSkinnedHandle
from ._scene_handles import PointCloudHandle as PointCloudHandle
from ._scene_handles import PointLightHandle as PointLightHandle
from ._scene_handles import RectAreaLightHandle as RectAreaLightHandle
from ._scene_handles import SceneNodeHandle as SceneNodeHandle
from ._scene_handles import SceneNodePointerEvent as SceneNodePointerEvent
from ._scene_handles import ScenePointerEvent as ScenePointerEvent
from ._scene_handles import SplineCatmullRomHandle as SplineCatmullRomHandle
from ._scene_handles import SplineCubicBezierHandle as SplineCubicBezierHandle
from ._scene_handles import SpotLightHandle as SpotLightHandle
from ._scene_handles import TransformControlsEvent as TransformControlsEvent
from ._scene_handles import TransformControlsHandle as TransformControlsHandle
from ._viser import CameraHandle as CameraHandle
from ._viser import ClientHandle as ClientHandle
from ._viser import InitialCameraConfig as InitialCameraConfig
from ._viser import ViserServer as ViserServer

__version__ = "1.0.24"



from typing import TYPE_CHECKING, Tuple, TypeVar, Union, cast

import numpy as onp

if TYPE_CHECKING:
    from .._base import MatrixLieGroup


T = TypeVar("T", bound="MatrixLieGroup")


def get_epsilon(dtype: onp.dtype) -> float:
    """Helper for grabbing type-specific precision constants.

    Args:
        dtype: Datatype.

    Returns:
        Output float.
    """
    if dtype == onp.float32:
        return 1e-5
    elif dtype == onp.float64:
        return 1e-10
    else:
        assert False


TupleOfBroadcastable = TypeVar(
    "TupleOfBroadcastable",
    bound="Tuple[Union[MatrixLieGroup, onp.ndarray], ...]",
)


def broadcast_leading_axes(inputs: TupleOfBroadcastable) -> TupleOfBroadcastable:
    """Broadcast leading axes of arrays. Takes tuples of either:
    - an array, which we assume has shape (*, D).
    - a Lie group object."""

    from .._base import MatrixLieGroup

    array_inputs = [
        (
            (x.parameters(), (x.parameters_dim,))
            if isinstance(x, MatrixLieGroup)
            else (x, x.shape[-1:])
        )
        for x in inputs
    ]
    for array, shape_suffix in array_inputs:
        assert array.shape[-len(shape_suffix) :] == shape_suffix
    batch_axes = onp.broadcast_shapes(
        *[array.shape[: -len(suffix)] for array, suffix in array_inputs]
    )
    broadcasted_arrays = tuple(
        onp.broadcast_to(array, batch_axes + shape_suffix)
        for (array, shape_suffix) in array_inputs
    )
    return cast(
        TupleOfBroadcastable,
        tuple(
            array if not isinstance(inp, MatrixLieGroup) else type(inp)(array)
            for array, inp in zip(broadcasted_arrays, inputs)
        ),
    )



"""TypeScript to Python TypedDict generator with AST-based parsing.

This script uses principled recursive descent parsing to convert TypeScript
definition files into Python TypedDict format with proper type mappings.

Key features:
- Full lexer and recursive descent parser for TypeScript types
- AST-based type conversion with semantic aliases
- Proper handling of complex union types and function signatures
- Conservative approach to undefined type references
"""

from __future__ import annotations

import re
import subprocess
from abc import ABC
from collections import defaultdict, deque
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

# ==================== TypeScript Type Parser ====================


class TokenType(Enum):
    # Literals
    IDENTIFIER = auto()
    STRING_LITERAL = auto()
    NUMBER_LITERAL = auto()

    # Operators and punctuation
    PIPE = auto()  # |
    AMPERSAND = auto()  # &
    QUESTION = auto()  # ?
    COLON = auto()  # :
    SEMICOLON = auto()  # ;
    COMMA = auto()  # ,
    DOT = auto()  # .
    ARROW = auto()  # =>

    # Brackets and parentheses
    LPAREN = auto()  # (
    RPAREN = auto()  # )
    LBRACKET = auto()  # [
    RBRACKET = auto()  # ]
    LBRACE = auto()  # {
    RBRACE = auto()  # }

    # Special
    EOF = auto()
    UNKNOWN = auto()


@dataclass
class Token:
    type: TokenType
    value: str
    position: int


class Lexer:
    """Tokenizes TypeScript type strings."""

    def __init__(self, text: str):
        self.text = text
        self.position = 0
        self.current_char = self.text[0] if text else None

    def advance(self):
        """Move to the next character."""
        self.position += 1
        if self.position >= len(self.text):
            self.current_char = None
        else:
            self.current_char = self.text[self.position]

    def peek(self, offset: int = 1) -> Optional[str]:
        """Look ahead at the next character without advancing."""
        peek_pos = self.position + offset
        if peek_pos >= len(self.text):
            return None
        return self.text[peek_pos]

    def skip_whitespace(self):
        """Skip whitespace characters."""
        while self.current_char is not None and self.current_char.isspace():
            self.advance()

    def read_string_literal(self) -> str:
        """Read a quoted string literal."""
        quote_char = self.current_char  # ' or "
        result = quote_char or ""
        self.advance()

        while self.current_char is not None and self.current_char != quote_char:
            if self.current_char == "\\":
                result += "\\"
                self.advance()
                if self.current_char is not None:
                    result += self.current_char
                    self.advance()
            else:
                result += self.current_char
                self.advance()

        if self.current_char == quote_char:
            result += self.current_char or ""
            self.advance()

        return result

    def read_number_literal(self) -> str:
        """Read a number literal."""
        result = ""
        while self.current_char is not None and (
            self.current_char.isdigit() or self.current_char in ".-"
        ):
            result += self.current_char
            self.advance()
        return result

    def read_identifier(self) -> str:
        """Read an identifier or keyword."""
        result = ""
        while self.current_char is not None and (
            self.current_char.isalnum() or self.current_char in "_$"
        ):
            result += self.current_char
            self.advance()
        return result

    def get_next_token(self) -> Token:
        """Get the next token from the input."""
        while self.current_char is not None:
            if self.current_char.isspace():
                self.skip_whitespace()
                continue

            # String literals
            if self.current_char in ['"', "'"]:
                return Token(
                    TokenType.STRING_LITERAL, self.read_string_literal(), self.position
                )

            # Number literals
            peek_char = self.peek()
            if self.current_char.isdigit() or (
                self.current_char == "-"
                and peek_char is not None
                and peek_char.isdigit()
            ):
                return Token(
                    TokenType.NUMBER_LITERAL, self.read_number_literal(), self.position
                )

            # Arrow function =>
            if self.current_char == "=" and self.peek() == ">":
                self.advance()
                self.advance()
                return Token(TokenType.ARROW, "=>", self.position - 2)

            # Single character tokens
            single_char_tokens = {
                "|": TokenType.PIPE,
                "&": TokenType.AMPERSAND,
                "?": TokenType.QUESTION,
                ":": TokenType.COLON,
                ";": TokenType.SEMICOLON,
                ",": TokenType.COMMA,
                ".": TokenType.DOT,
                "(": TokenType.LPAREN,
                ")": TokenType.RPAREN,
                "[": TokenType.LBRACKET,
                "]": TokenType.RBRACKET,
                "{": TokenType.LBRACE,
                "}": TokenType.RBRACE,
            }

            if self.current_char in single_char_tokens:
                token_type = single_char_tokens[self.current_char]
                char = self.current_char
                pos = self.position
                self.advance()
                return Token(token_type, char, pos)

            # Identifiers and keywords
            if self.current_char.isalpha() or self.current_char in "_$":
                return Token(
                    TokenType.IDENTIFIER, self.read_identifier(), self.position
                )

            # Unknown character
            char = self.current_char
            pos = self.position
            self.advance()
            return Token(TokenType.UNKNOWN, char, pos)

        return Token(TokenType.EOF, "", self.position)

    def tokenize(self) -> List[Token]:
        """Tokenize the entire input string."""
        tokens = []
        while True:
            token = self.get_next_token()
            tokens.append(token)
            if token.type == TokenType.EOF:
                break
        return tokens


# AST Node classes
class TypeNode(ABC):
    """Base class for all type AST nodes."""

    pass


@dataclass
class PrimitiveType(TypeNode):
    """Primitive TypeScript types like string, number, boolean."""

    name: str


@dataclass
class LiteralType(TypeNode):
    """Literal types like "hello", 42, true."""

    value: str
    literal_type: str  # 'string', 'number', 'boolean'


@dataclass
class IdentifierType(TypeNode):
    """Type identifiers like MyInterface, HTMLElement."""

    name: str


@dataclass
class UnionType(TypeNode):
    """Union types like string | number."""

    types: List[TypeNode]


@dataclass
class IntersectionType(TypeNode):
    """Intersection types like A & B."""

    types: List[TypeNode]


@dataclass
class ArrayType(TypeNode):
    """Array types like string[] or Array<T>."""

    element_type: TypeNode


@dataclass
class TupleType(TypeNode):
    """Tuple types like [string, number]."""

    element_types: List[TypeNode]


@dataclass
class FunctionType(TypeNode):
    """Function types like (x: number) => string."""

    parameters: List["ParameterType"]
    return_type: TypeNode


@dataclass
class ParameterType(TypeNode):
    """Function parameter like x: number or x?: string."""

    name: Optional[str]
    type: TypeNode
    optional: bool = False


@dataclass
class ParenthesizedType(TypeNode):
    """Parenthesized type like (string | number)."""

    inner_type: TypeNode


@dataclass
class GenericType(TypeNode):
    """Generic type like Array<T> or Map<K, V>."""

    base_type: TypeNode
    type_arguments: List[TypeNode]


class TypeScriptTypeParser:
    """Recursive descent parser for TypeScript type annotations."""

    def __init__(self, tokens: List[Token]):
        self.tokens = tokens
        self.position = 0
        self.current_token = self.tokens[0] if tokens else Token(TokenType.EOF, "", 0)

    def advance(self):
        """Move to the next token."""
        if self.position < len(self.tokens) - 1:
            self.position += 1
            self.current_token = self.tokens[self.position]

    def peek(self, offset: int = 1) -> Token:
        """Look ahead at a future token."""
        peek_pos = self.position + offset
        if peek_pos < len(self.tokens):
            return self.tokens[peek_pos]
        return Token(TokenType.EOF, "", 0)

    def parse(self) -> TypeNode:
        """Parse the token stream into an AST."""
        return self.parse_union_type()

    def parse_union_type(self) -> TypeNode:
        """Parse union type: A | B | C."""
        left = self.parse_intersection_type()

        if self.current_token.type == TokenType.PIPE:
            types = [left]
            while self.current_token.type == TokenType.PIPE:
                self.advance()  # consume |
                types.append(self.parse_intersection_type())
            return UnionType(types)

        return left

    def parse_intersection_type(self) -> TypeNode:
        """Parse intersection type: A & B & C."""
        left = self.parse_array_type()

        if self.current_token.type == TokenType.AMPERSAND:
            types = [left]
            while self.current_token.type == TokenType.AMPERSAND:
                self.advance()  # consume &
                types.append(self.parse_array_type())
            return IntersectionType(types)

        return left

    def parse_array_type(self) -> TypeNode:
        """Parse array type: T[] or Array<T>."""
        base = self.parse_primary_type()

        # Handle T[]
        while self.current_token.type == TokenType.LBRACKET:
            self.advance()  # consume [
            if self.current_token.type == TokenType.RBRACKET:
                self.advance()  # consume ]
                base = ArrayType(base)
            else:
                # This might be a tuple type [T, U] - backtrack and parse as primary
                break

        return base

    def parse_primary_type(self) -> TypeNode:
        """Parse primary types: identifiers, literals, parenthesized, etc."""

        # Parenthesized type or function type
        if self.current_token.type == TokenType.LPAREN:
            return self.parse_parenthesized_or_function()

        # Tuple type [T, U, V]
        if self.current_token.type == TokenType.LBRACKET:
            return self.parse_tuple_type()

        # String literal
        if self.current_token.type == TokenType.STRING_LITERAL:
            value = self.current_token.value
            self.advance()
            return LiteralType(value, "string")

        # Number literal
        if self.current_token.type == TokenType.NUMBER_LITERAL:
            value = self.current_token.value
            self.advance()
            return LiteralType(value, "number")

        # Identifier (including primitive types)
        if self.current_token.type == TokenType.IDENTIFIER:
            name = self.current_token.value
            self.advance()

            # Check for qualified names like HTMLElement.prototype
            if self.current_token.type == TokenType.DOT:
                qualified_parts = [name]
                while self.current_token.type == TokenType.DOT:
                    self.advance()  # consume .
                    if self.current_token.type == TokenType.IDENTIFIER:
                        qualified_parts.append(self.current_token.value)
                        self.advance()
                    else:
                        break
                return IdentifierType(".".join(qualified_parts))

            # Check if it's a primitive type
            if name in [
                "string",
                "number",
                "boolean",
                "void",
                "null",
                "undefined",
                "any",
                "unknown",
                "never",
            ]:
                return PrimitiveType(name)

            return IdentifierType(name)

        # If we can't parse anything, create an unknown identifier
        return IdentifierType("unknown")

    def parse_parenthesized_or_function(self) -> TypeNode:
        """Parse parenthesized type (T) or function type (params) => return."""
        self.advance()  # consume (

        # Check if this looks like a function parameter list
        if self.is_function_parameter_list():
            return self.parse_function_type_from_paren()

        # Otherwise, parse as parenthesized type
        inner = self.parse_union_type()

        if self.current_token.type == TokenType.RPAREN:
            self.advance()  # consume )

        # Check for arrow function after parentheses
        if self.current_token.type == TokenType.ARROW:
            self.advance()  # consume =>
            return_type = self.parse_union_type()
            # Convert the parenthesized type to function parameters if possible
            params = []
            if isinstance(inner, IdentifierType):
                params = [ParameterType(None, inner)]
            return FunctionType(params, return_type)

        return ParenthesizedType(inner)

    def is_function_parameter_list(self) -> bool:
        """Check if the current position looks like a function parameter list."""
        # Look ahead to see if we have parameter-like patterns
        saved_pos = self.position

        try:
            # Skip to see if we find => after )
            paren_depth = 1  # We already consumed the opening (
            while paren_depth > 0 and self.current_token.type != TokenType.EOF:
                if self.current_token.type == TokenType.LPAREN:
                    paren_depth += 1
                elif self.current_token.type == TokenType.RPAREN:
                    paren_depth -= 1
                self.advance()

            # Now check if we have =>
            result = self.current_token.type == TokenType.ARROW
            return result
        finally:
            # Restore position
            self.position = saved_pos
            self.current_token = (
                self.tokens[self.position]
                if self.position < len(self.tokens)
                else Token(TokenType.EOF, "", 0)
            )

    def parse_function_type_from_paren(self) -> FunctionType:
        """Parse function type starting from after the opening parenthesis."""
        parameters = []

        # Parse parameter list
        while self.current_token.type not in {TokenType.RPAREN, TokenType.EOF}:
            param = self.parse_parameter()
            parameters.append(param)

            if self.current_token.type == TokenType.COMMA:
                self.advance()  # consume ,
            elif self.current_token.type == TokenType.RPAREN:
                break
            else:
                # Safety break: if we don't see comma or closing paren, advance to avoid infinite loop
                self.advance()

        if self.current_token.type == TokenType.RPAREN:
            self.advance()  # consume )

        # Expect =>
        if self.current_token.type == TokenType.ARROW:
            self.advance()  # consume =>

        # Parse return type
        return_type = self.parse_union_type()

        return FunctionType(parameters, return_type)

    def parse_parameter(self) -> ParameterType:
        """Parse a function parameter."""
        name = None
        optional = False

        # Try to parse name: type pattern
        if self.current_token.type == TokenType.IDENTIFIER and self.peek().type in [
            TokenType.COLON,
            TokenType.QUESTION,
        ]:
            name = self.current_token.value
            self.advance()

            if self.current_token.type == TokenType.QUESTION:
                optional = True
                self.advance()  # consume ?

            if self.current_token.type == TokenType.COLON:
                self.advance()  # consume :

        # Parse the type
        param_type = self.parse_union_type()

        return ParameterType(name, param_type, optional)

    def parse_tuple_type(self) -> TypeNode:
        """Parse tuple type [T, U, V]."""
        self.advance()  # consume [

        element_types = []

        while self.current_token.type not in {TokenType.RBRACKET, TokenType.EOF}:
            # Handle labeled tuple elements like [min: number, max: number]
            if (
                self.current_token.type == TokenType.IDENTIFIER
                and self.peek().type == TokenType.COLON
            ):
                # Skip the label
                self.advance()  # identifier
                self.advance()  # :

            element_type = self.parse_union_type()
            element_types.append(element_type)

            if self.current_token.type == TokenType.COMMA:
                self.advance()  # consume ,
            elif self.current_token.type == TokenType.RBRACKET:
                break
            else:
                # Safety break: if we don't see comma or closing bracket, advance to avoid infinite loop
                self.advance()

        if self.current_token.type == TokenType.RBRACKET:
            self.advance()  # consume ]

        return TupleType(element_types)


def parse_typescript_type(type_string: str) -> TypeNode:
    """Parse a TypeScript type string into an AST."""
    lexer = Lexer(type_string)
    tokens = lexer.tokenize()
    parser = TypeScriptTypeParser(tokens)
    return parser.parse()


# ==================== AST to Python Converter ====================


class PythonTypeConverter:
    """Converts TypeScript AST nodes to Python type annotations."""

    def __init__(
        self,
        known_interfaces: Optional[Set[str]] = None,
        type_aliases: Optional[Dict[str, str]] = None,
        enums: Optional[Dict[str, "TypeScriptEnum"]] = None,
        use_literal_enums: bool = True,
    ):
        self.known_interfaces = known_interfaces or set()
        self.type_aliases = type_aliases or {}
        self.enums = enums or {}
        self.use_literal_enums = use_literal_enums

        # Semantic type mappings
        self.primitive_mappings = {
            "string": "str",
            "number": "float",
            "boolean": "bool",
            "void": "None",
            "null": "None",
            "undefined": "None",
            "any": "Any",
            "unknown": "Any",
            "never": "Never",
        }

        # DOM and browser type mappings
        self.dom_types = {
            "HTMLElement",
            "Element",
            "Node",
            "Event",
            "MouseEvent",
            "DOMRect",
            "HTMLCanvasElement",
            "HTMLDivElement",
            "HTMLInputElement",
            "EventTarget",
            "Document",
            "Window",
        }

        # CSS and canvas types
        self.css_types = {
            "CSSStyleDeclaration",
            "CanvasRenderingContext2D",
            "CanvasGradient",
            "CanvasPattern",
            "ImageData",
            "TextMetrics",
        }

    def convert(self, node: TypeNode) -> str:
        """Convert a TypeScript AST node to Python type annotation."""
        if isinstance(node, PrimitiveType):
            return self._convert_primitive(node)
        elif isinstance(node, LiteralType):
            return self._convert_literal(node)
        elif isinstance(node, IdentifierType):
            return self._convert_identifier(node)
        elif isinstance(node, UnionType):
            return self._convert_union(node)
        elif isinstance(node, IntersectionType):
            return self._convert_intersection(node)
        elif isinstance(node, ArrayType):
            return self._convert_array(node)
        elif isinstance(node, TupleType):
            return self._convert_tuple(node)
        elif isinstance(node, FunctionType):
            return self._convert_function(node)
        elif isinstance(node, ParenthesizedType):
            return self.convert(node.inner_type)
        elif isinstance(node, GenericType):
            return self._convert_generic(node)
        else:
            return "UnknownType"

    def _convert_primitive(self, node: PrimitiveType) -> str:
        """Convert primitive types."""
        return self.primitive_mappings.get(node.name, "UnknownType")

    def _convert_literal(self, node: LiteralType) -> str:
        """Convert literal types."""
        return f"Literal[{node.value}]"

    def _convert_identifier(self, node: IdentifierType) -> str:
        """Convert identifier types with semantic mapping."""
        name = node.name

        # Handle qualified names (e.g., HTMLElement.prototype)
        if "." in name:
            parts = name.split(".")
            base_name = parts[0]

            # Check for type aliases first (e.g., Scale.Auto, Sync.Scales)
            # Try exact match first.
            if name in self.type_aliases:
                alias_def = self.type_aliases[name]
                # Recursively convert the aliased type
                try:
                    ast_node = parse_typescript_type(alias_def)
                    return self.convert(ast_node)
                except Exception:
                    # If parsing fails, fall back to original logic
                    pass

            # Try with common namespace prefixes (e.g., Sync.Scales -> Cursor.Sync.Scales)
            for prefix in ["Cursor", "Scale", "Series", "Axis", "Legend"]:
                qualified_name = f"{prefix}.{name}"
                if qualified_name in self.type_aliases:
                    alias_def = self.type_aliases[qualified_name]
                    try:
                        ast_node = parse_typescript_type(alias_def)
                        return self.convert(ast_node)
                    except Exception:
                        continue

            # Check for namespace patterns first (e.g., Legend.Markers -> Legend_Markers)
            full_name = "_".join(parts)
            if full_name in self.known_interfaces:
                return full_name

            # Check for namespaced enums (e.g., Axis.Align -> Align)
            enum_name = parts[-1]  # Get the last part (e.g., "Align" from "Axis.Align")
            if enum_name in self.enums:
                if self.use_literal_enums:
                    return self._convert_enum_to_literal(enum_name)
                else:
                    return enum_name
            elif enum_name in self.known_interfaces:
                return enum_name

            # For type aliases within namespaces (e.g., Legend.Width, Legend.Stroke)
            # these should be treated as UnknownType since they're not interfaces
            if base_name in self.known_interfaces:
                return "UnknownType"

            # For other qualified names that we don't have definitions for
            return "UnknownType"

        # Check if it's a type alias first
        if name in self.type_aliases:
            alias_def = self.type_aliases[name]
            # Recursively convert the aliased type
            try:
                ast_node = parse_typescript_type(alias_def)
                return self.convert(ast_node)
            except Exception:
                # If parsing fails, fall back to original logic
                pass

        # Check if it's an enum first
        if name in self.enums:
            if self.use_literal_enums:
                return self._convert_enum_to_literal(name)
            else:
                return name

        # Check if it's a known interface
        if name in self.known_interfaces:
            return name

        # Check for DOM types
        if name in self.dom_types:
            return "DOMElement"

        # Check for CSS/Canvas types
        if name in self.css_types:
            return "CSSValue"

        # Handle common TypeScript utility types
        utility_types = {
            "Partial",
            "Required",
            "Readonly",
            "Pick",
            "Omit",
            "Exclude",
            "Extract",
            "NonNullable",
            "Parameters",
            "ConstructorParameters",
            "ReturnType",
            "InstanceType",
            "ThisType",
            "Record",
            "Uppercase",
            "Lowercase",
            "Capitalize",
            "Uncapitalize",
        }
        if name in utility_types:
            return "UnknownType"

        # Handle built-in TypeScript types
        builtin_types = {
            "object": "Dict",
            "Object": "Dict",
            "String": "str",
            "Number": "float",
            "Boolean": "bool",
            "Array": "List",
            "Function": "JSCallback",
            "Date": "UnknownType",
            "RegExp": "UnknownType",
            "Promise": "UnknownType",
            "Symbol": "UnknownType",
            "BigInt": "int",
        }
        if name in builtin_types:
            return builtin_types[name]

        # Default for unknown identifiers
        return "UnknownType"

    def _convert_enum_to_literal(self, enum_name: str) -> str:
        """Convert an enum to a Literal union type."""
        if enum_name not in self.enums:
            return "UnknownType"

        enum_obj = self.enums[enum_name]
        values = []
        for member_name, value in enum_obj.members:
            # Keep the value as-is (already properly quoted for strings)
            values.append(value)

        # Use compact Literal syntax: Literal[0, 1, 2] instead of Literal[0] | Literal[1] | Literal[2]
        return f"Literal[{', '.join(values)}]"

    def _convert_union(self, node: UnionType) -> str:
        """Convert union types."""
        if not node.types:
            return "UnknownType"

        converted_types = []
        for type_node in node.types:
            converted = self.convert(type_node)
            converted_types.append(converted)

        # Remove duplicates while preserving order
        seen = set()
        unique_types = []
        for t in converted_types:
            if t not in seen:
                seen.add(t)
                unique_types.append(t)

        if len(unique_types) == 1:
            return unique_types[0]
        elif len(unique_types) > 1:
            return f"Union[{', '.join(unique_types)}]"
        else:
            return "UnknownType"

    def _convert_intersection(self, node: IntersectionType) -> str:
        """Convert intersection types - these are complex in Python."""
        # For now, just return the first type or UnknownType
        if node.types:
            return self.convert(node.types[0])
        return "UnknownType"

    def _convert_array(self, node: ArrayType) -> str:
        """Convert array types to Tuple[T, ...] by default."""
        element_type = self.convert(node.element_type)
        return f"Tuple[{element_type}, ...]"

    def _convert_tuple(self, node: TupleType) -> str:
        """Convert tuple types."""
        if not node.element_types:
            return "Tuple[()]"

        element_types = [self.convert(t) for t in node.element_types]
        return f"Tuple[{', '.join(element_types)}]"

    def _convert_function(self, node: FunctionType) -> str:
        """Convert function types to JSCallback."""
        return "JSCallback"

    def _convert_generic(self, node: GenericType) -> str:
        """Convert generic types."""
        base = self.convert(node.base_type)

        # Handle common generic types
        if isinstance(node.base_type, IdentifierType):
            base_name = node.base_type.name

            if base_name == "Array" and node.type_arguments:
                element_type = self.convert(node.type_arguments[0])
                return f"Tuple[{element_type}, ...]"

            if base_name == "Record" and len(node.type_arguments) >= 2:
                key_type = self.convert(node.type_arguments[0])
                value_type = self.convert(node.type_arguments[1])
                return f"Dict[{key_type}, {value_type}]"

            if base_name == "Partial":
                # Partial<T> in TypeScript makes all properties optional
                # In Python TypedDict, we handle this at the TypedDict level
                return "UnknownType"

            if base_name == "Promise":
                # Promises don't have a direct Python equivalent
                return "UnknownType"

        # For other generic types, just use the base type
        return base


def convert_type_with_ast(
    ts_type: str,
    known_interfaces: Set[str],
    type_aliases: Optional[Dict[str, str]] = None,
    enums: Optional[Dict[str, "TypeScriptEnum"]] = None,
    use_literal_enums: bool = True,
) -> str:
    """Convert TypeScript type using AST-based parsing."""
    try:
        # Create converter with known interfaces, type aliases, and enums for better type resolution
        converter = PythonTypeConverter(
            known_interfaces, type_aliases, enums, use_literal_enums
        )

        # Parse and convert using AST
        ast_node = parse_typescript_type(ts_type)
        return converter.convert(ast_node)
    except Exception:
        # Fallback for unparseable types
        return "UnknownType"


# ==================== TypeScript Interface Parser ====================


@dataclass
class Interface:
    name: str
    properties: List[Dict[str, str]]
    raw_definition: str
    base_interface: Optional[str] = None


@dataclass
class TypeAlias:
    name: str
    definition: str


@dataclass
class TypeScriptEnum:
    name: str
    members: List[tuple[str, str]]  # List of (member_name, value) pairs.


def parse_type_aliases(content: str) -> Dict[str, str]:
    """Parse type alias declarations from TypeScript content with proper namespace tracking."""
    type_aliases = {}
    lines = content.split("\n")
    i = 0
    namespace_stack = []
    namespace_brace_depth = []  # Track brace depth for each namespace.
    current_brace_depth = 0

    while i < len(lines):
        line = lines[i].strip()

        # Count braces to track nesting depth.
        open_braces = line.count("{")
        close_braces = line.count("}")

        # Track namespace entries.
        namespace_match = re.match(r"(export |declare )?namespace (\w+)", line)
        if namespace_match:
            namespace_name = namespace_match.group(2)
            namespace_stack.append(namespace_name)
            # After this line, we'll be one level deeper.
            namespace_brace_depth.append(current_brace_depth + open_braces)

        # Update current brace depth.
        current_brace_depth += open_braces - close_braces

        # Check if we've exited any namespaces.
        while namespace_stack and current_brace_depth < namespace_brace_depth[-1]:
            namespace_stack.pop()
            namespace_brace_depth.pop()

        # Find type alias declarations: export type Show = boolean | ...
        type_match = re.match(r"export type (\w+)\s*=\s*(.+)", line)
        if type_match:
            type_name = type_match.group(1)
            type_def = type_match.group(2).rstrip(";")

            # Build qualified name with namespace (skip main uPlot namespace).
            relevant_namespaces = [ns for ns in namespace_stack if ns != "uPlot"]
            if relevant_namespaces:
                qualified_name = ".".join(relevant_namespaces + [type_name])
                type_aliases[qualified_name] = type_def
                # Also add unqualified name for fallback.
                type_aliases[type_name] = type_def
            else:
                type_aliases[type_name] = type_def

        i += 1

    return type_aliases


def parse_enums(content: str) -> Dict[str, TypeScriptEnum]:
    """Parse const enum declarations from TypeScript content."""
    enums = {}
    lines = content.split("\n")
    i = 0

    while i < len(lines):
        line = lines[i].strip()

        # Find const enum declarations.
        enum_match = re.match(r"export const enum (\w+)", line)
        if enum_match:
            enum_name = enum_match.group(1)

            # Extract enum body.
            enum_members = []
            i += 1

            while i < len(lines):
                current_line = lines[i].strip()

                if current_line == "}":
                    break

                # Parse enum member: Name = value (either number or string).
                # Handle both numeric and string values
                member_match = re.match(r"(\w+)\s*=\s*(.+?)(?:,.*)?$", current_line)
                if member_match:
                    member_name = member_match.group(1)
                    value = member_match.group(2).strip().rstrip(",")
                    enum_members.append((member_name, value))

                i += 1

            if enum_members:
                enums[enum_name] = TypeScriptEnum(name=enum_name, members=enum_members)

        i += 1

    return enums


def parse_interfaces(content: str) -> List[Interface]:
    """Parse interfaces from TypeScript content with namespace support."""
    interfaces = []
    lines = content.split("\n")
    i = 0
    namespace_stack = []

    while i < len(lines):
        line = lines[i].strip()

        # Track namespace entries.
        namespace_match = re.match(r"(export |declare )?namespace (\w+)", line)
        if namespace_match:
            namespace_name = namespace_match.group(2)
            namespace_stack.append(namespace_name)
            i += 1
            continue

        # Track namespace exits (simple heuristic).
        if line == "}" and namespace_stack:
            indent = len(lines[i]) - len(lines[i].lstrip())
            # If we're at low indentation, probably exiting a namespace.
            if indent <= len(namespace_stack):
                namespace_stack.pop()

        # Find interface declarations (both exported and internal).
        interface_match = re.match(r"(export )?interface (\w+)", line)
        if interface_match:
            interface_name = interface_match.group(2)

            # Build qualified name with namespace (skip main uPlot namespace for top-level exports).
            if namespace_stack and not (
                len(namespace_stack) == 1 and namespace_stack[0] == "uPlot"
            ):
                qualified_name = "_".join(namespace_stack + [interface_name])
            else:
                qualified_name = interface_name

            # Check for inheritance using extends keyword
            extends_match = re.search(r"extends\s+([A-Za-z_][A-Za-z0-9_.<>]*)", line)
            base_interface = None
            if extends_match:
                base_interface = extends_match.group(1)

            # Extract interface body.
            body_lines = []
            brace_count = 0
            started = False

            while i < len(lines):
                current_line = lines[i]

                if "{" in current_line:
                    started = True

                if started:
                    body_lines.append(current_line)
                    brace_count += current_line.count("{") - current_line.count("}")
                    if brace_count == 0:
                        break

                i += 1

            # Parse properties.
            properties = parse_properties("\n".join(body_lines))

            interfaces.append(
                Interface(
                    name=qualified_name,
                    properties=properties,
                    raw_definition="\n".join(body_lines),
                    base_interface=base_interface,
                )
            )

        i += 1

    return interfaces


def parse_properties(interface_body: str) -> List[Dict[str, str]]:
    """Parse properties from interface body."""
    properties = []
    lines = interface_body.split("\n")

    i = 0
    while i < len(lines):
        line = lines[i].strip()

        if not line or line in ["{", "}"]:
            i += 1
            continue

        # Collect comments.
        comment = ""
        while (
            line.startswith("/**")
            or line.startswith("//")
            or (line.startswith("*") and ":" not in line)
        ):
            if line.startswith("/**"):
                comment = line[3:].strip()
            elif line.startswith("//"):
                comment = line[2:].strip()
            elif line.startswith("*"):
                comment = line[1:].strip()

            if comment.endswith("*/"):
                comment = comment[:-2].strip()

            i += 1
            line = lines[i].strip() if i < len(lines) else ""

        # Parse index signatures like [key: string]: Scale.
        if line.startswith("[") and "]:" in line:
            # Extract index signature: [key: string]: Scale -> key_type: string, value_type: Scale.
            match = re.match(r"^\[([^:]+):\s*([^\]]+)\]:\s*(.+);?$", line.strip())
            if match:
                key_name = match.group(1).strip()
                key_type = match.group(2).strip()
                value_type = match.group(3).strip().rstrip(";").rstrip(",")

                # Create a special property to indicate this is an index signature.
                properties.append(
                    {
                        "name": "__index_signature__",
                        "type": f"[{key_type}, {value_type}]",  # Store both key and value types.
                        "optional": False,
                        "comment": f"Index signature: {key_name} -> {value_type}",
                    }
                )

        # Parse regular property line.
        elif ":" in line and not line.startswith("["):
            # Handle multi-line properties - but be careful about parsing.
            prop_line = line

            # Only continue to next line if this line doesn't end with ; or ,.
            while (
                not prop_line.rstrip().endswith(";")
                and not prop_line.rstrip().endswith(",")
                and i + 1 < len(lines)
            ):
                i += 1
                next_line = lines[i].strip()
                # Stop if next line looks like a new property or comment.
                if (
                    ":" in next_line
                    or next_line.startswith("/**")
                    or next_line.startswith("//")
                ):
                    i -= 1  # Back up.
                    break
                prop_line += " " + next_line

            prop = parse_property(prop_line, comment)
            if prop:
                properties.append(prop)

        i += 1

    return properties


def parse_property(
    line: str, comment: str = ""
) -> Optional[Dict[str, Union[str, bool]]]:
    """Parse a single property line."""
    line = line.strip().rstrip(";").rstrip(",")

    if ":" not in line:
        return None

    optional = "?" in line
    parts = line.split(":", 1)
    if len(parts) != 2:
        return None

    name = parts[0].replace("?", "").strip()
    type_str = parts[1].strip()

    # Clean up type string - remove everything after comma or comment.
    if "," in type_str:
        type_str = type_str.split(",")[0].strip()

    # Remove inline comments.
    if "//" in type_str:
        type_str = type_str.split("//")[0].strip()

    if "/*" in type_str:
        type_str = type_str.split("/*")[0].strip()

    # Handle readonly.
    if name.startswith("readonly "):
        name = name[9:]

    # Validate property name.
    if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", name):
        return None

    return {"name": name, "type": type_str, "optional": optional, "comment": comment}


def camel_to_screaming_snake(name: str) -> str:
    """Convert camelCase to SCREAMING_SNAKE_CASE for enum members."""
    # Insert underscores before capital letters (except the first)
    result = re.sub(r"(?<!^)(?=[A-Z])", "_", name)
    return result.upper()


def topological_sort_interfaces(interfaces: List[Interface]) -> List[Interface]:
    """Sort interfaces in topological order so dependencies come before dependents."""
    interface_map = {iface.name: iface for iface in interfaces}
    interface_names = set(interface_map.keys())

    # Build dependency graph: interface -> set of interfaces it depends on.
    dependencies = defaultdict(set)
    dependents = defaultdict(set)  # Reverse mapping.

    for interface in interfaces:
        interface_name = interface.name
        deps = set()

        # Find all interface dependencies in properties.
        for prop in interface.properties:
            # Skip index signatures.
            if prop["name"] == "__index_signature__":
                continue

            prop_type = prop["type"]

            # Find interface references in the type string.
            # Look for exact interface names (including underscore names like Legend_Markers).
            for other_name in interface_names:
                if other_name != interface_name:
                    # Use word boundary regex to ensure exact matches.
                    # But exclude patterns like "Legend.Something" which are type aliases, not interface refs.
                    if re.search(r"\b" + re.escape(other_name) + r"\b", prop_type):
                        # Make sure it's not a qualified name like "Legend.Width".
                        # By checking if it's followed by a dot and identifier.
                        full_match = re.search(
                            r"\b"
                            + re.escape(other_name)
                            + r"(\.[A-Za-z_][A-Za-z0-9_]*)?",
                            prop_type,
                        )
                        if full_match and full_match.group(1):
                            # This is a qualified name like "Legend.Width", not an interface reference.
                            continue
                        deps.add(other_name)

            # Also look for namespaced references that map to underscore names.
            # e.g., "Legend.Markers" should create dependency on "Legend_Markers".
            namespaced_refs = re.findall(
                r"\b[A-Z][a-zA-Z0-9_]*\.[A-Z][a-zA-Z0-9_]*\b", prop_type
            )
            for ref in namespaced_refs:
                underscore_name = ref.replace(".", "_")
                if (
                    underscore_name in interface_names
                    and underscore_name != interface_name
                ):
                    deps.add(underscore_name)

        dependencies[interface_name] = deps
        for dep in deps:
            dependents[dep].add(interface_name)

    # Kahn's algorithm for topological sorting.
    # Start with interfaces that have no dependencies.
    in_degree = {iface.name: len(dependencies[iface.name]) for iface in interfaces}
    queue = deque([name for name, degree in in_degree.items() if degree == 0])
    result = []

    while queue:
        current = queue.popleft()
        result.append(interface_map[current])

        # Remove this interface from its dependents' dependency lists.
        for dependent in dependents[current]:
            in_degree[dependent] -= 1
            if in_degree[dependent] == 0:
                queue.append(dependent)

    # Check for cycles.
    if len(result) != len(interfaces):
        remaining = set(interface_map.keys()) - {iface.name for iface in result}
        print(f"Warning: Circular dependencies detected in interfaces: {remaining}")
        # Add remaining interfaces in original order.
        for interface in interfaces:
            if interface.name in remaining:
                result.append(interface)

    return result


def resolve_inheritance(interfaces: List[Interface]) -> List[Interface]:
    """Resolve interface inheritance by copying properties from base interfaces."""
    interface_map = {iface.name: iface for iface in interfaces}
    resolved_interfaces = []

    # Create a simple resolution that doesn't cause loops.
    for interface in interfaces:
        if interface.base_interface:
            # Look for the base interface with proper namespacing.
            base_name = interface.base_interface
            base_interface = None

            # Try with namespace prefix if the current interface has one.
            current_namespace = ""
            if "_" in interface.name:
                current_namespace = interface.name.split("_")[0] + "_"

            # Try exact match first.
            if base_name in interface_map:
                base_interface = interface_map[base_name]
            else:
                namespaced_base = current_namespace + base_name
                if namespaced_base in interface_map:
                    base_interface = interface_map[namespaced_base]

            if base_interface:
                # Recursively resolve base interface first if it has inheritance.
                resolved_base = base_interface
                if base_interface.base_interface:
                    # Find the base's base.
                    base_base_name = base_interface.base_interface
                    base_base_interface = None

                    if base_base_name in interface_map:
                        base_base_interface = interface_map[base_base_name]
                    else:
                        namespaced_base_base = current_namespace + base_base_name
                        if namespaced_base_base in interface_map:
                            base_base_interface = interface_map[namespaced_base_base]

                    if base_base_interface:
                        # Combine base's base properties with base properties.
                        combined_base_props = base_base_interface.properties.copy()
                        base_prop_names = {
                            prop["name"] for prop in base_interface.properties
                        }
                        combined_base_props = [
                            prop
                            for prop in combined_base_props
                            if prop["name"] not in base_prop_names
                        ]
                        combined_base_props.extend(base_interface.properties)

                        resolved_base = Interface(
                            name=base_interface.name,
                            properties=combined_base_props,
                            raw_definition=base_interface.raw_definition,
                            base_interface=None,
                        )

                # Copy properties from resolved base interface.
                combined_properties = resolved_base.properties.copy()

                # Add properties from current interface (they override base properties).
                current_prop_names = {prop["name"] for prop in interface.properties}
                combined_properties = [
                    prop
                    for prop in combined_properties
                    if prop["name"] not in current_prop_names
                ]
                combined_properties.extend(interface.properties)

                resolved_interface = Interface(
                    name=interface.name,
                    properties=combined_properties,
                    raw_definition=interface.raw_definition,
                    base_interface=None,  # Clear base interface after resolution.
                )
                resolved_interfaces.append(resolved_interface)
            else:
                # Base interface not found, keep as is.
                resolved_interfaces.append(interface)
        else:
            # No inheritance, keep as is.
            resolved_interfaces.append(interface)

    return resolved_interfaces


def find_needed_interfaces(
    interfaces: List[Interface], start: str = "Options"
) -> Set[str]:
    """Find interfaces needed starting from the given interface."""
    needed = set()
    to_process = [start]
    interface_map = {iface.name: iface for iface in interfaces}

    while to_process:
        current = to_process.pop()
        if current in needed or current not in interface_map:
            continue

        needed.add(current)
        current_interface = interface_map[current]

        # Also include base interfaces for inheritance
        if current_interface.base_interface:
            base_name = current_interface.base_interface
            # Try exact match first.
            if base_name in interface_map and base_name not in needed:
                to_process.append(base_name)
            else:
                # Try with namespace prefix.
                current_namespace = ""
                if "_" in current:
                    current_namespace = current.split("_")[0] + "_"

                namespaced_base = current_namespace + base_name
                if namespaced_base in interface_map and namespaced_base not in needed:
                    to_process.append(namespaced_base)

        # Look for interface references in properties.
        for prop in current_interface.properties:
            # Find PascalCase words that might be interface names.
            words = re.findall(r"\b[A-Z][a-zA-Z0-9_]*\b", prop["type"])
            for word in words:
                if word in interface_map and word not in needed:
                    to_process.append(word)

            # Also look for namespaced interfaces like Axis.Grid -> Axis_Grid.
            namespaced_refs = re.findall(
                r"\b[A-Z][a-zA-Z0-9_]*\.[A-Z][a-zA-Z0-9_]*\b", prop["type"]
            )
            for ref in namespaced_refs:
                underscore_name = ref.replace(".", "_")
                if underscore_name in interface_map and underscore_name not in needed:
                    to_process.append(underscore_name)

    return needed


def collect_used_enums(
    interfaces: List[Interface],
    enums: Dict[str, TypeScriptEnum],
    type_aliases: Dict[str, str],
) -> Set[str]:
    """Collect enum names that are actually used in the TypedDict definitions."""
    used_enums = set()

    # Create set of known interface names for type conversion
    known_interface_names = {iface.name for iface in interfaces} | set(enums.keys())

    for interface in interfaces:
        for prop in interface.properties:
            if prop["name"] == "__index_signature__":
                continue

            # Convert the property type and see what enums it references (using enum classes for analysis).
            converted_type = convert_type_with_ast(
                prop["type"],
                known_interface_names,
                type_aliases,
                enums,
                use_literal_enums=False,
            )

            # Look for enum references in the converted type.
            for enum_name in enums.keys():
                if re.search(rf"\b{re.escape(enum_name)}\b", converted_type):
                    used_enums.add(enum_name)

    return used_enums


def generate_python_code(
    interfaces: List[Interface], content: str, use_literal_enums: bool = True
) -> str:
    """Generate clean Python TypedDict code using AST-based type conversion."""

    # Find needed interfaces.
    needed_names = find_needed_interfaces(interfaces, "Options")
    needed_interfaces = [iface for iface in interfaces if iface.name in needed_names]

    # Resolve inheritance relationships
    needed_interfaces = resolve_inheritance(needed_interfaces)

    # Filter out width and height from Options interface.
    for interface in needed_interfaces:
        if interface.name == "Options":
            interface.properties = [
                prop
                for prop in interface.properties
                if prop["name"] not in ["width", "height"]
            ]

    # Sort interfaces in topological order to eliminate forward references.
    needed_interfaces = topological_sort_interfaces(needed_interfaces)

    # Parse enums and type aliases from content.
    all_enums = parse_enums(content)
    type_aliases = parse_type_aliases(content)

    # Only keep enums that are actually used in the final interfaces (if generating enum classes).
    if use_literal_enums:
        # When using literal enums, we still need all enums for conversion but won't generate classes.
        enums = all_enums
    else:
        used_enum_names = collect_used_enums(needed_interfaces, all_enums, type_aliases)
        enums = {
            name: enum_obj
            for name, enum_obj in all_enums.items()
            if name in used_enum_names
        }

    # Create set of known interface names for better type conversion.
    known_interface_names = {iface.name for iface in needed_interfaces} | set(
        enums.keys()
    )

    lines = [
        '"""Clean uPlot TypedDict definitions."""',
        "",
        "# =========================================================================",
        "# WARNING: This file is auto-generated by _generate_types.py",
        "# DO NOT EDIT MANUALLY - Your changes will be overwritten!",
        "# To modify types, edit the generation script or the TypeScript definitions",
        "# =========================================================================",
        "",
        "from __future__ import annotations",
        "",
    ]

    # Add imports based on what we're generating.
    if enums and not use_literal_enums:
        lines.append("from enum import IntEnum, StrEnum")
    lines.extend(
        [
            "from typing import Any, Dict, List, Literal, Tuple, Union",
            "from typing_extensions import Never, Required, TypedDict",
            "",
            "# Semantic type aliases for unsupported/complex TypeScript patterns",
            "JSCallback = Never  # JavaScript function signatures",
            "DOMElement = Never  # DOM elements (HTMLElement, etc.)",
            "CSSValue = str  # CSS property values",
            "UnknownType = Any  # Unknown interface references",
            "",
        ]
    )

    # Generate Enum classes and type aliases (only if not using literal enums).
    if enums and not use_literal_enums:
        lines.append("# Enum definitions")
        for enum_name, enum_obj in enums.items():
            # Determine if this is a string enum or numeric enum.
            is_string_enum = any(
                value.startswith("'") or value.startswith('"')
                for _, value in enum_obj.members
            )

            if is_string_enum:
                lines.append(f"class {enum_name}(StrEnum):")
                for member_name, value in enum_obj.members:
                    # For string enums, use UPPER_CASE naming convention.
                    upper_name = member_name.upper()
                    lines.append(f"    {upper_name} = {value}")
            else:
                lines.append(f"class {enum_name}(IntEnum):")
                for member_name, value in enum_obj.members:
                    snake_case_name = camel_to_screaming_snake(member_name)
                    lines.append(f"    {snake_case_name} = {value}")
            lines.append("")

    # Generate TypedDict classes and collect type aliases.
    type_aliases_to_generate = []
    interface_names = {iface.name for iface in needed_interfaces}

    # First pass: collect type aliases that need to be generated.
    for interface in needed_interfaces:
        name = interface.name
        properties = interface.properties

        # Check for index signatures.
        index_signatures = [p for p in properties if p["name"] == "__index_signature__"]
        regular_properties = [
            p for p in properties if p["name"] != "__index_signature__"
        ]

        # Handle interfaces with only index signatures (like Scales).
        if index_signatures and not regular_properties:
            # Store for generation before classes.
            index_sig = index_signatures[0]
            # Parse [key_type, value_type] from the stored format.
            match = re.match(r"^\[([^,]+),\s*(.+)\]$", index_sig["type"])
            if match:
                key_type = convert_type_with_ast(
                    match.group(1).strip(),
                    known_interface_names,
                    type_aliases,
                    enums,
                    use_literal_enums,
                )
                value_type = convert_type_with_ast(
                    match.group(2).strip(),
                    known_interface_names,
                    type_aliases,
                    enums,
                    use_literal_enums,
                )
                type_aliases_to_generate.append(
                    (name, index_sig["comment"], key_type, value_type)
                )

    # Generate type aliases first (using string quotes for forward references).
    if type_aliases_to_generate:
        lines.append("# Type aliases for index signatures")
        for name, comment, key_type, value_type in type_aliases_to_generate:
            # Use string quotes to allow forward references.
            value_type_str = (
                f'"{value_type}"' if value_type in interface_names else value_type
            )
            # Add period to comment if it doesn't already end with punctuation.
            if comment and not comment.rstrip().endswith((".", "!", "?", ":", ";")):
                comment += "."
            lines.extend(
                [f"# {comment}", f"{name} = Dict[{key_type}, {value_type_str}]", ""]
            )

    # Second pass: generate TypedDict classes.
    type_alias_names = {name for name, _, _, _ in type_aliases_to_generate}

    for interface in needed_interfaces:
        name = interface.name

        # Skip if this was already handled as a type alias.
        if name in type_alias_names:
            continue

        properties = interface.properties

        # Check for index signatures.
        index_signatures = [p for p in properties if p["name"] == "__index_signature__"]
        regular_properties = [
            p for p in properties if p["name"] != "__index_signature__"
        ]

        if not regular_properties:
            lines.append(f"{name} = TypedDict('{name}', {{}}, total=False)")
            lines.append('""":py:class:`~typing.TypedDict` type for uPlot."""')
            lines.append("")
            continue

        # Separate required and optional regular properties.
        required = [p for p in regular_properties if not p["optional"]]
        optional = [p for p in regular_properties if p["optional"]]

        # Generate functional TypedDict.
        lines.append(f"{name} = TypedDict('{name}', {{")

        # Add all properties (both required and optional).
        all_props = required + optional
        for i, prop in enumerate(all_props):
            prop_name = prop["name"]  # Use original name, no escaping needed in quotes.

            # Use AST-based type conversion.
            prop_type = convert_type_with_ast(
                prop["type"],
                known_interface_names,
                type_aliases,
                enums,
                use_literal_enums,
            )

            # Add comment above the property.
            if prop["comment"]:
                comment = prop["comment"]
                # Add period if comment doesn't already end with punctuation.
                if comment and not comment.rstrip().endswith((".", "!", "?", ":", ";")):
                    comment += "."
                lines.append(f"    # {comment}")

            # Handle required vs optional.
            if required and optional:  # Mixed case - use Required[] for required props.
                if prop in required:
                    type_annotation = f"Required[{prop_type}]"
                else:
                    type_annotation = prop_type
            else:
                # All required or all optional.
                type_annotation = prop_type

            # Add trailing comma except for last item.
            comma = "," if i < len(all_props) - 1 else ""
            lines.append(f"    '{prop_name}': {type_annotation}{comma}")

        # Set total=False if we have any optional properties.
        total_param = ", total=False" if optional else ""
        lines.append(f"}}{total_param})")

        # Add docstring after the TypedDict definition.
        lines.append('""":py:class:`~typing.TypedDict` type for uPlot."""')
        lines.append("")

    return "\n".join(lines)


def main(
    input_file: Path = Path("uPlot.d.ts"),
    output_file: Path = Path("_uplot_types.py"),
    use_literal_enums: bool = True,
) -> None:
    """Generate clean Python TypedDict types from TypeScript definitions."""

    print(f"Reading {input_file}")
    content = input_file.read_text()

    print("Parsing interfaces...")
    interfaces = parse_interfaces(content)
    print(f"Found {len(interfaces)} interfaces")

    print("Generating Python code...")
    enum_mode = "Literal types" if use_literal_enums else "Enum classes"
    print(f"Using {enum_mode} for enums")
    python_code = generate_python_code(interfaces, content, use_literal_enums)

    print(f"Writing to {output_file}")
    output_file.write_text(python_code)

    print("Running ruff format...")
    try:
        subprocess.run(["ruff", "format", str(output_file)], check=True)
        print("Formatting completed successfully")
    except subprocess.CalledProcessError as e:
        print(f"Warning: ruff format failed with exit code {e.returncode}")
    except FileNotFoundError:
        print("Warning: ruff not found, skipping formatting")

    print("Done!")


@dataclass
class Config:
    """Configuration for generating Python TypedDict types from TypeScript definitions."""

    input: Path = Path("uPlot.d.ts")
    """Input TypeScript file"""

    output: Path = Path("_uplot_types.py")
    """Output Python file"""

    enum_classes: bool = False
    """Generate enum classes instead of Literal types"""


if __name__ == "__main__":
    import tyro

    config = tyro.cli(Config)

    main(
        input_file=config.input,
        output_file=config.output,
        use_literal_enums=not config.enum_classes,
    )



from typing import Union

import numpy as np
import numpy.typing as npt

# Type aliases Numpy arrays; primarily for function inputs.

Scalar = Union[float, npt.NDArray[np.floating]]
"""Type alias for `Union[float, Array]`."""


__all__ = [
    "Scalar",
]
